import logging
import os
import posixpath
import tempfile
from string import Template

import yaml

from snowflake import snowpark
from snowflake.ml._internal import file_utils
from snowflake.ml._internal.utils import identifier
from snowflake.ml.model import _model_meta
from snowflake.ml.model._deploy_client.image_builds import (
    base_image_builder,
    client_image_builder,
    docker_context,
)
from snowflake.ml.model._deploy_client.utils import constants, snowservice_client

logger = logging.getLogger(__name__)


class ServerImageBuilder(base_image_builder.ImageBuilder):
    """
    Server-side image building and upload to model registry.
    """

    def __init__(
        self,
        *,
        id: str,
        image_repo: str,
        model_meta: _model_meta.ModelMetadata,
        session: snowpark.Session,
        artifact_stage_location: str,
        compute_pool: str,
    ) -> None:
        """Initialization

        Args:
            id: A hexadecimal string used for naming the image tag.
            image_repo: Path to image repository.
            model_meta: Model Metadata.
            session: Snowpark session
            artifact_stage_location: Spec file and future deployment related artifacts will be stored under
                {stage}/models/{model_id}
            compute_pool: The compute pool used to run docker image build workload.
        """
        self.model_id = id
        self.image_repo = image_repo
        self.image_tag = "/".join([image_repo.rstrip("/"), id]) + ":latest"
        self.model_meta = model_meta
        self.session = session
        self.artifact_stage_location = artifact_stage_location
        self.compute_pool = compute_pool
        self.client = snowservice_client.SnowServiceClient(session)

        assert artifact_stage_location.startswith(
            "@"
        ), f"stage path should start with @, actual: {artifact_stage_location}"

    def build_and_upload_image(self) -> str:
        """
        Builds and uploads an image to the model registry.
        """
        logger.info("Starting server-side image build with Kaniko")
        with tempfile.TemporaryDirectory() as context_dir:
            dc = docker_context.DockerContext(context_dir=context_dir, model_meta=self.model_meta)
            dc.build()
            self._build_image_in_remote_job(context_dir)
        return self.image_tag

    def _build_image_in_remote_job(self, context_dir: str) -> None:
        """
        Args:
            context_dir: Path to context directory.

        """
        context_tarball_stage_location = f"{self.artifact_stage_location}/{constants.CONTEXT}.tar.gz"
        spec_stage_location = f"{self.artifact_stage_location}/{constants.IMAGE_BUILD_JOB_SPEC_TEMPLATE}.yaml"
        kaniko_shell_script_stage_location = f"{self.artifact_stage_location}/{constants.KANIKO_SHELL_SCRIPT_NAME}"

        self._compress_and_upload_docker_context_tarball(
            context_dir=context_dir, context_tarball_stage_location=context_tarball_stage_location
        )

        self._construct_and_upload_docker_entrypoint_script(
            context_dir=context_dir, context_tarball_stage_location=context_tarball_stage_location
        )

        # This is more of a workaround to support non-spcs-registry images.
        # TODO[shchen] remove such logic when first-party-image is supported on snowservice registry.
        # The regular Kaniko image doesn't include a shell; only the debug image comes with a shell. We need a shell
        # as we use an sh script to launch Kaniko
        kaniko_image_tag = "/".join([self.image_repo.rstrip("/"), "kaniko-project/executor:debug"])
        image_builder_client = client_image_builder.ClientImageBuilder(
            id=self.model_id,
            image_repo=self.image_repo,
            image_tag=kaniko_image_tag,
            model_meta=self.model_meta,
            session=self.session,
        )
        base_image = image_builder_client.build_and_upload_image(image_to_pull="gcr.io/kaniko-project/executor:debug")

        self._construct_and_upload_job_spec(
            base_image=base_image,
            context_dir=context_dir,
            kaniko_shell_script_stage_location=kaniko_shell_script_stage_location,
        )
        self._launch_kaniko_job(spec_stage_location)

    def _construct_and_upload_docker_entrypoint_script(
        self, context_dir: str, context_tarball_stage_location: str
    ) -> None:
        """Construct a shell script that invokes logic to uncompress the docker context tarball, then invoke Kaniko
        executor to build images and push to image registry; the script will also ensure the docker credential(used to
        authenticate to image registry) stays up-to-date when session token refreshes.

        Args:
            context_dir: Path to context directory.
            context_tarball_stage_location: Path context directory stage location.
        """

        kaniko_shell_script_template = os.path.join(
            os.path.dirname(__file__), f"templates/{constants.KANIKO_SHELL_SCRIPT_TEMPLATE}"
        )
        kaniko_shell_file = os.path.join(context_dir, constants.KANIKO_SHELL_SCRIPT_NAME)

        with open(kaniko_shell_script_template, encoding="utf-8") as template_file, open(
            kaniko_shell_file, "w+", encoding="utf-8"
        ) as script_file:
            normed_artifact_stage_path = posixpath.normpath(identifier.remove_prefix(self.artifact_stage_location, "@"))
            params = {
                # Remove @ in the beginning, append "/" to denote root directory.
                "tar_from": "/" + posixpath.normpath(identifier.remove_prefix(context_tarball_stage_location, "@")),
                # Remove @ in the beginning, append "/" to denote root directory.
                "tar_to": "/" + normed_artifact_stage_path,
                "context_dir": f"dir:///{normed_artifact_stage_path}/{constants.CONTEXT}",
                "image_repo": self.image_repo,
                # All models will be sharing the same layer cache from the image_repo/cache directory.
                "cache_repo": f"{self.image_repo.rstrip('/')}/cache",
                "image_destination": self.image_tag,
            }
            template = Template(template_file.read())
            script = template.safe_substitute(params)
            script_file.write(script)
            logger.debug(f"script content: \n\n {script}")
        self.session.file.put(
            local_file_name=kaniko_shell_file,
            stage_location=self.artifact_stage_location,
            auto_compress=False,
            overwrite=True,
        )

    def _compress_and_upload_docker_context_tarball(
        self, context_dir: str, context_tarball_stage_location: str
    ) -> None:
        try:
            with file_utils._create_tar_gz_stream(source_dir=context_dir, arcname=constants.CONTEXT) as input_stream:
                self.session.file.put_stream(
                    input_stream=input_stream,
                    stage_location=context_tarball_stage_location,
                    auto_compress=False,
                    overwrite=True,
                )
        except Exception as e:
            raise RuntimeError(
                "Exception occurred when compressing docker context dir as tarball and upload to stage", e
            )

    def _construct_and_upload_job_spec(
        self, base_image: str, context_dir: str, kaniko_shell_script_stage_location: str
    ) -> None:
        assert kaniko_shell_script_stage_location.startswith(
            "@"
        ), f"stage path should start with @, actual: {kaniko_shell_script_stage_location}"
        spec_template_path = os.path.join(
            os.path.dirname(__file__), f"templates/{constants.IMAGE_BUILD_JOB_SPEC_TEMPLATE}"
        )
        spec_file_path = os.path.join(os.path.dirname(context_dir), f"{constants.IMAGE_BUILD_JOB_SPEC_TEMPLATE}.yaml")

        with open(spec_template_path, encoding="utf-8") as template_file, open(
            spec_file_path, "w+", encoding="utf-8"
        ) as spec_file:
            assert self.artifact_stage_location.startswith("@")
            normed_artifact_stage_path = posixpath.normpath(identifier.remove_prefix(self.artifact_stage_location, "@"))
            (db, schema, stage, path) = identifier.parse_schema_level_object_identifier(normed_artifact_stage_path)
            content = Template(template_file.read()).substitute(
                {
                    "base_image": base_image,
                    "container_name": constants.KANIKO_CONTAINER_NAME,
                    "stage": identifier.get_schema_level_object_identifier(db, schema, stage),
                    # Remove @ in the beginning, append "/" to denote root directory.
                    "script_path": "/"
                    + posixpath.normpath(identifier.remove_prefix(kaniko_shell_script_stage_location, "@")),
                }
            )
            content_dict = yaml.safe_load(content)
            yaml.dump(content_dict, spec_file)
            spec_file.seek(0)
            logger.debug(f"Kaniko job spec file: \n\n {spec_file.read()}")

        self.session.file.put(
            local_file_name=spec_file_path,
            stage_location=self.artifact_stage_location,
            auto_compress=False,
            overwrite=True,
        )

    def _launch_kaniko_job(self, spec_stage_location: str) -> None:
        job_id = self.client.create_job(compute_pool=self.compute_pool, spec_stage_location=spec_stage_location)
        logger.debug(f"Submit job for building docker image in kaniko with job id {job_id}")
        # Given image build can take a while, we set a generous timeout to be 1 hour.
        self.client.block_until_resource_is_ready(
            resource_name=job_id,
            resource_type=constants.ResourceType.JOB,
            container_name=constants.KANIKO_CONTAINER_NAME,
            max_retries=240,
            retry_interval_secs=15,
        )



from typing import List, Tuple

# TODO(amauser): Move this scheme and registry creation in general into a server-side implementation.
_REGISTRY_TABLE_SCHEMA: List[Tuple[str, str]] = [
    ("CREATION_CONTEXT", "VARCHAR"),
    ("CREATION_ENVIRONMENT_SPEC", "OBJECT"),
    ("CREATION_ROLE", "VARCHAR"),
    ("CREATION_TIME", "TIMESTAMP_TZ"),
    ("ID", "VARCHAR PRIMARY KEY RELY"),
    ("INPUT_SPEC", "OBJECT"),
    ("NAME", "VARCHAR"),
    ("OUTPUT_SPEC", "OBJECT"),
    ("RUNTIME_ENVIRONMENT_SPEC", "OBJECT"),
    ("TRAINING_DATASET_ID", "VARCHAR"),
    ("TYPE", "VARCHAR"),
    ("URI", "VARCHAR"),
    ("VERSION", "VARCHAR"),
]

# TODO(amauser): Generalize attribute to any column reference.
_METADATA_TABLE_SCHEMA: List[Tuple[str, str]] = [
    ("ATTRIBUTE_NAME", "VARCHAR"),
    ("EVENT_ID", "VARCHAR UNIQUE NOT NULL"),
    ("EVENT_TIMESTAMP", "TIMESTAMP_TZ"),
    ("MODEL_ID", "VARCHAR FOREIGN KEY REFERENCES {registry_table_name}(ID) RELY"),
    ("OPERATION", "VARCHAR"),
    ("ROLE", "VARCHAR"),
    ("SEQUENCE_ID", "BIGINT AUTOINCREMENT START 0 INCREMENT 1 PRIMARY KEY"),
    ("VALUE", "OBJECT"),
]

_DEPLOYMENTS_TABLE_SCHEMA: List[Tuple[str, str]] = [
    ("CREATION_TIME", "TIMESTAMP_TZ"),
    ("MODEL_ID", "VARCHAR FOREIGN KEY REFERENCES {registry_table_name}(ID) RELY"),
    ("DEPLOYMENT_NAME", "VARCHAR"),
    ("OPTIONS", "VARIANT"),
    ("TARGET_PLATFORM", "VARCHAR"),
    ("ROLE", "VARCHAR"),
    ("STAGE_PATH", "VARCHAR"),
    ("SIGNATURE", "VARIANT"),
    ("TARGET_METHOD", "VARCHAR"),
]

_ARTIFACT_TABLE_SCHEMA: List[Tuple[str, str]] = [
    ("ID", "VARCHAR"),
    ("TYPE", "VARCHAR"),
    ("NAME", "VARCHAR"),
    ("VERSION", "VARCHAR"),
    ("CREATION_ROLE", "VARCHAR"),
    ("CREATION_TIME", "TIMESTAMP_TZ"),
    ("ARTIFACT_SPEC", "OBJECT"),
    # Below is out-of-line constraints of Snowflake table.
    # See https://docs.snowflake.com/en/sql-reference/sql/create-table
    ("PRIMARY KEY", "(ID, TYPE) RELY"),
]



import collections
import copy
import re
import textwrap
import warnings
from importlib import metadata as importlib_metadata
from typing import DefaultDict, Dict, List, Optional, Tuple

from packaging import requirements, specifiers, utils as packaging_utils, version

import snowflake.connector
from snowflake.ml._internal import env as snowml_env
from snowflake.ml._internal.utils import query_result_checker
from snowflake.snowpark import session

_SNOWML_PKG_NAME = "snowflake-ml-python"
_INFO_SCHEMA_PACKAGES_HAS_RUNTIME_VERSION: Optional[bool] = None
_SNOWFLAKE_CONDA_PACKAGE_CACHE: Dict[str, List[version.Version]] = {}

DEFAULT_CHANNEL_NAME = ""


def _validate_pip_requirement_string(req_str: str) -> requirements.Requirement:
    """Validate the input pip requirement string according to PEP 508.

    Args:
        req_str: The string contains the pip requirement specification.

    Raises:
        ValueError: Raised when python is specified as a dependency.
        ValueError: Raised when invalid requirement string confronted.
        ValueError: Raised when requirement containing marker record confronted.

    Returns:
        A requirements.Requirement object containing the requirement information.
    """
    try:
        r = requirements.Requirement(req_str)
        r.name = packaging_utils.canonicalize_name(r.name)

        if r.name == "python":
            raise ValueError("Don't specify python as a dependency, use python version argument instead.")
    except requirements.InvalidRequirement:
        raise ValueError(f"Invalid package requirement {req_str} found.")

    if r.marker:
        raise ValueError("Markers is not supported in conda dependency.")
    return r


def _validate_conda_dependency_string(dep_str: str) -> Tuple[str, requirements.Requirement]:
    """Validate conda dependency string like `pytorch == 1.12.1` or `conda-forge::transformer` and split the channel
        name before the double colon and requirement specification after that.

    Args:
        dep_str: The string contains the conda dependency specification.

    Raises:
        ValueError: Raised when invalid conda dependency containing extras record confronted.
        ValueError: Raised when invalid conda dependency containing url record confronted.
        ValueError: Raised when conda dependency operator ~= which is not supported by conda.

    Returns:
        A tuple containing the conda channel name and requirement.Requirement object showing requirement information.
    """
    channel_str, _, requirement_str = dep_str.rpartition("::")
    r = _validate_pip_requirement_string(requirement_str)
    if channel_str != "pip":
        if r.extras:
            raise ValueError("Extras is not supported in conda dependency.")
        if r.url:
            raise ValueError("Url is not supported in conda dependency.")
        for specifier in r.specifier:
            if specifier.operator == "~=":
                raise ValueError("Operator ~= is not supported in conda dependency.")
    return (channel_str, r)


def _check_if_requirement_same(req_a: requirements.Requirement, req_b: requirements.Requirement) -> bool:
    """Check if two requirements are the same package.

    Args:
        req_a: The first requirement.
        req_b: The second requirement.

    Returns:
        Whether two requirements should and need to be merged.
    """
    return req_a.name == req_b.name


class DuplicateDependencyError(Exception):
    ...


class DuplicateDependencyInMultipleChannelsError(Exception):
    ...


def append_requirement_list(req_list: List[requirements.Requirement], p_req: requirements.Requirement) -> None:
    """Append a requirement to an existing requirement list. If need and able to merge, merge it, otherwise, append it.

    Args:
        req_list: The target requirement list.
        p_req: The requirement to append.

    Raises:
        DuplicateDependencyError: Raised when the same package being added.
    """
    for req in req_list:
        if _check_if_requirement_same(p_req, req):
            raise DuplicateDependencyError(
                f"Found duplicate dependencies in pip requirements: {str(req)} and {str(p_req)}."
            )
    req_list.append(p_req)


def append_conda_dependency(
    conda_chan_deps: DefaultDict[str, List[requirements.Requirement]], p_chan_dep: Tuple[str, requirements.Requirement]
) -> None:
    """Append a conda dependency to an existing conda dependencies dict, if not existed in any channel.
        To avoid making unnecessary modification to dict, we check the existence first, then try to merge, then append,
        if still needed.

    Args:
        conda_chan_deps: The target dependencies dict.
        p_chan_dep: The tuple of channel and dependency to append.

    Raises:
        DuplicateDependencyError: Raised when the same package required from one channel.
        DuplicateDependencyInMultipleChannelsError: Raised when the same package required from different channels.
    """
    p_channel, p_req = p_chan_dep
    for chan_channel, chan_req_list in conda_chan_deps.items():
        for chan_req in chan_req_list:
            if _check_if_requirement_same(p_req, chan_req):
                if chan_channel != p_channel:
                    raise DuplicateDependencyInMultipleChannelsError(
                        "Found duplicate dependencies: "
                        + f"{str(chan_req)} in channel {chan_channel} and {str(p_req)} in channel {p_channel}."
                    )
                else:
                    raise DuplicateDependencyError(
                        f"Found duplicate dependencies in channel {chan_channel}: {str(chan_req)} and {str(p_req)}."
                    )
    conda_chan_deps[p_channel].append(p_req)


def validate_pip_requirement_string_list(req_str_list: List[str]) -> List[requirements.Requirement]:
    """Validate the a list of pip requirement string according to PEP 508.

    Args:
        req_str_list: The list of string contains the pip requirement specification.

    Returns:
        A requirements.Requirement list containing the requirement information.
    """
    seen_pip_requirement_list: List[requirements.Requirement] = []
    for req_str in req_str_list:
        append_requirement_list(seen_pip_requirement_list, _validate_pip_requirement_string(req_str=req_str))

    return seen_pip_requirement_list


def validate_conda_dependency_string_list(dep_str_list: List[str]) -> DefaultDict[str, List[requirements.Requirement]]:
    """Validate a list of conda dependency string, find any duplicate package across different channel and create a dict
        to represent the whole dependencies.

    Args:
        dep_str_list: The list of string contains the conda dependency specification.

    Returns:
        A dict mapping from the channel name to the list of requirements from that channel.
    """
    validated_conda_dependency_list = list(map(_validate_conda_dependency_string, dep_str_list))
    ret_conda_dependency_dict: DefaultDict[str, List[requirements.Requirement]] = collections.defaultdict(list)
    for p_channel, p_req in validated_conda_dependency_list:
        append_conda_dependency(ret_conda_dependency_dict, (p_channel, p_req))

    return ret_conda_dependency_dict


def get_local_installed_version_of_pip_package(pip_req: requirements.Requirement) -> requirements.Requirement:
    """Get the local installed version of a given pip package requirement.
        If the package is locally installed, and the local version meet the specifier of the requirements, return a new
        requirement specifier that pins the version.
        If the local version does not meet the specifier of the requirements, a warn will be omitted and the original
        requirement specifier is returned.
        If the package is not locally installed or not found, the original package requirement is returned.

    Args:
        pip_req: A requirements.Requirement object showing the requirement.

    Returns:
        A requirements.Requirement object that might have version pinned to local installed version.
    """
    try:
        local_dist = importlib_metadata.distribution(pip_req.name)
        local_dist_version = local_dist.version
    except importlib_metadata.PackageNotFoundError:
        if pip_req.name == _SNOWML_PKG_NAME:
            local_dist_version = snowml_env.VERSION
        else:
            return pip_req
    if pip_req.specifier.contains(local_dist_version):
        new_pip_req = copy.deepcopy(pip_req)
        new_pip_req.specifier = specifiers.SpecifierSet(specifiers=f"=={local_dist_version}")
        return new_pip_req
    else:
        warnings.warn(
            f"Package requirement {str(pip_req)} specified, while version {local_dist_version} is installed.",
            category=UserWarning,
        )
        return pip_req


def relax_requirement_version(req: requirements.Requirement) -> requirements.Requirement:
    """Remove version specifier from a requirement.

    Args:
        req: The requirement that version specifier to be removed.

    Returns:
        A new requirement object without version specifier while others kept.
    """
    new_req = copy.deepcopy(req)
    new_req.specifier = specifiers.SpecifierSet()
    return new_req


def _check_runtime_version_column_existence(session: session.Session) -> bool:
    sql = textwrap.dedent(
        """
        SHOW COLUMNS
        LIKE 'runtime_version'
        IN TABLE information_schema.packages;
        """
    )
    result = session.sql(sql).count()
    return result == 1


def validate_requirements_in_snowflake_conda_channel(
    session: session.Session, reqs: List[requirements.Requirement], python_version: str
) -> Optional[List[str]]:
    """Search the snowflake anaconda channel for packages with version meet the specifier.

    Args:
        session: Snowflake connection session.
        reqs: List of requirement specifiers.
        python_version: A string of python version where model is run.

    Raises:
        ValueError: Raised when the specifier cannot be supported when creating UDF.

    Returns:
        A list of pinned latest version that available in Snowflake anaconda channel and meet the version specifier.
    """
    global _INFO_SCHEMA_PACKAGES_HAS_RUNTIME_VERSION

    if _INFO_SCHEMA_PACKAGES_HAS_RUNTIME_VERSION is None:
        _INFO_SCHEMA_PACKAGES_HAS_RUNTIME_VERSION = _check_runtime_version_column_existence(session)
    ret_list = []
    reqs_to_request = []
    for req in reqs:
        if req.name not in _SNOWFLAKE_CONDA_PACKAGE_CACHE:
            reqs_to_request.append(req)
    if reqs_to_request:
        pkg_names_str = " OR ".join(
            f"package_name = '{req_name}'" for req_name in sorted(req.name for req in reqs_to_request)
        )
        if _INFO_SCHEMA_PACKAGES_HAS_RUNTIME_VERSION:
            parsed_python_version = version.Version(python_version)
            sql = textwrap.dedent(
                f"""
                SELECT PACKAGE_NAME, VERSION
                FROM information_schema.packages
                WHERE ({pkg_names_str})
                AND language = 'python'
                AND (runtime_version = '{parsed_python_version.major}.{parsed_python_version.minor}'
                    OR runtime_version is null);
                """
            )
        else:
            sql = textwrap.dedent(
                f"""
                SELECT PACKAGE_NAME, VERSION
                FROM information_schema.packages
                WHERE ({pkg_names_str})
                AND language = 'python';
                """
            )

        try:
            result = (
                query_result_checker.SqlResultValidator(
                    session=session,
                    query=sql,
                )
                .has_column("VERSION")
                .has_dimensions(expected_rows=None, expected_cols=2)
                .validate()
            )
            for row in result:
                req_name = row["PACKAGE_NAME"]
                req_ver = version.parse(row["VERSION"])
                cached_req_ver_list = _SNOWFLAKE_CONDA_PACKAGE_CACHE.get(req_name, [])
                cached_req_ver_list.append(req_ver)
                _SNOWFLAKE_CONDA_PACKAGE_CACHE[req_name] = cached_req_ver_list
        except snowflake.connector.DataError:
            return None
    for req in reqs:
        if len(req.specifier) > 1 or any(spec.operator != "==" for spec in req.specifier):
            raise ValueError("At most 1 version specifier using == operator is supported without local conda resolver.")
        available_versions = list(req.specifier.filter(set(_SNOWFLAKE_CONDA_PACKAGE_CACHE.get(req.name, []))))
        if not available_versions:
            return None
        else:
            ret_list.append(str(req))
    return sorted(ret_list)


# We have to use re to support MLFlow generated python string, which use = rather than ==
PYTHON_VERSION_PATTERN = re.compile(r"python(?:(?P<op>=|==|>|<|>=|<=|~=|===)(?P<ver>\d(?:\.\d+)+))?")


def parse_python_version_string(dep: str) -> Optional[str]:
    if dep.startswith("python"):
        m = PYTHON_VERSION_PATTERN.search(dep)
        if m is None:
            return None
        op = m.group("op")
        if op and (op != "=" and op != "=="):
            raise ValueError("Unsupported operator for python version specifier.")
        ver = m.group("ver")
        if ver:
            return ver
        else:
            # "python" only, no specifier
            return ""
    return None


def _find_conda_dep_spec(
    conda_chan_deps: DefaultDict[str, List[requirements.Requirement]], pkg_name: str
) -> Optional[Tuple[str, requirements.Requirement]]:
    for channel in conda_chan_deps:
        spec = next(filter(lambda req: req.name == pkg_name, conda_chan_deps[channel]), None)
        if spec:
            return channel, spec
    return None


def _find_pip_req_spec(pip_reqs: List[requirements.Requirement], pkg_name: str) -> Optional[requirements.Requirement]:
    spec = next(filter(lambda req: req.name == pkg_name, pip_reqs), None)
    return spec


def _find_dep_spec(
    conda_chan_deps: DefaultDict[str, List[requirements.Requirement]],
    pip_reqs: List[requirements.Requirement],
    conda_pkg_name: str,
    pip_pkg_name: Optional[str] = None,
    remove_spec: bool = False,
) -> Tuple[
    DefaultDict[str, List[requirements.Requirement]], List[requirements.Requirement], Optional[requirements.Requirement]
]:
    if pip_pkg_name is None:
        pip_pkg_name = conda_pkg_name
    spec_conda = _find_conda_dep_spec(conda_chan_deps, conda_pkg_name)
    if spec_conda:
        channel, spec = spec_conda
        if remove_spec:
            conda_chan_deps[channel].remove(spec)
        return conda_chan_deps, pip_reqs, spec
    else:
        spec_pip = _find_pip_req_spec(pip_reqs, pip_pkg_name)
        if spec_pip:
            if remove_spec:
                pip_reqs.remove(spec_pip)
            return conda_chan_deps, pip_reqs, spec_pip
    return conda_chan_deps, pip_reqs, None


def generate_env_for_cuda(
    conda_chan_deps: DefaultDict[str, List[requirements.Requirement]],
    pip_reqs: List[requirements.Requirement],
    cuda_version: str,
) -> Tuple[DefaultDict[str, List[requirements.Requirement]], List[requirements.Requirement]]:
    conda_chan_deps_cuda = copy.deepcopy(conda_chan_deps)
    pip_reqs_cuda = copy.deepcopy(pip_reqs)

    cuda_version_obj = version.parse(cuda_version)
    cuda_version_spec_str = f"{cuda_version_obj.major}.{cuda_version_obj.minor}.*"

    try:
        append_conda_dependency(
            conda_chan_deps_cuda,
            ("nvidia", requirements.Requirement(f"cuda=={cuda_version_spec_str}")),
        )
    except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
        pass

    conda_chan_deps_cuda, pip_reqs_cuda, xgboost_spec = _find_dep_spec(
        conda_chan_deps_cuda, pip_reqs, conda_pkg_name="xgboost", remove_spec=True
    )
    if xgboost_spec:
        xgboost_spec.name = "py-xgboost-gpu"
        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                ("conda-forge", xgboost_spec),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

    conda_chan_deps_cuda, pip_reqs_cuda, pytorch_spec = _find_dep_spec(
        conda_chan_deps_cuda, pip_reqs, conda_pkg_name="pytorch", pip_pkg_name="torch", remove_spec=True
    )
    if pytorch_spec:
        pytorch_spec.name = "pytorch"
        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                ("pytorch", pytorch_spec),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                p_chan_dep=("pytorch", requirements.Requirement(f"pytorch-cuda=={cuda_version_spec_str}")),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

    conda_chan_deps_cuda, pip_reqs_cuda, tf_spec = _find_dep_spec(
        conda_chan_deps_cuda, pip_reqs, conda_pkg_name="tensorflow", remove_spec=True
    )
    if tf_spec:
        tf_spec.name = "tensorflow-gpu"
        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                ("conda-forge", tf_spec),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

    conda_chan_deps_cuda, pip_reqs_cuda, transformers_spec = _find_dep_spec(
        conda_chan_deps_cuda, pip_reqs, conda_pkg_name="transformers", remove_spec=False
    )
    if transformers_spec:
        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                ("conda-forge", requirements.Requirement("accelerate>=0.22.0")),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

        # Required by bitsandbytes
        try:
            append_conda_dependency(
                conda_chan_deps_cuda,
                (DEFAULT_CHANNEL_NAME, get_local_installed_version_of_pip_package(requirements.Requirement("scipy"))),
            )
        except (DuplicateDependencyError, DuplicateDependencyInMultipleChannelsError):
            pass

        try:
            append_requirement_list(
                pip_reqs_cuda,
                requirements.Requirement("bitsandbytes>=0.41.0"),
            )
        except DuplicateDependencyError:
            pass

    return conda_chan_deps_cuda, pip_reqs_cuda



import os
from typing import TYPE_CHECKING, Callable, Optional, Type, Union, cast

import cloudpickle
import numpy as np
import pandas as pd
from typing_extensions import TypeGuard, Unpack

from snowflake.ml._internal import type_utils
from snowflake.ml.model import (
    _model_meta as model_meta_api,
    custom_model,
    model_signature,
    type_hints as model_types,
)
from snowflake.ml.model._handlers import _base
from snowflake.ml.model._signatures import numpy_handler, utils as model_signature_utils

if TYPE_CHECKING:
    import sklearn.base
    import sklearn.pipeline


class _SKLModelHandler(_base._ModelHandler[Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"]]):
    """Handler for scikit-learn based model.

    Currently sklearn.base.BaseEstimator and sklearn.pipeline.Pipeline based classes are supported.
    """

    handler_type = "sklearn"
    DEFAULT_TARGET_METHODS = ["predict", "transform", "predict_proba", "predict_log_proba", "decision_function"]

    @staticmethod
    def can_handle(
        model: model_types.SupportedModelType,
    ) -> TypeGuard[Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"]]:
        return (
            (
                type_utils.LazyType("sklearn.base.BaseEstimator").isinstance(model)
                or type_utils.LazyType("sklearn.pipeline.Pipeline").isinstance(model)
            )
            and (not type_utils.LazyType("xgboost.XGBModel").isinstance(model))  # XGBModel is actually a BaseEstimator
            and any(
                (hasattr(model, method) and callable(getattr(model, method, None)))
                for method in _SKLModelHandler.DEFAULT_TARGET_METHODS
            )
        )

    @staticmethod
    def cast_model(
        model: model_types.SupportedModelType,
    ) -> Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"]:
        import sklearn.base
        import sklearn.pipeline

        assert isinstance(model, sklearn.base.BaseEstimator) or isinstance(model, sklearn.pipeline.Pipeline)

        return cast(Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"], model)

    @staticmethod
    def _save_model(
        name: str,
        model: Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"],
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        sample_input: Optional[model_types.SupportedDataType] = None,
        is_sub_model: Optional[bool] = False,
        **kwargs: Unpack[model_types.SKLModelSaveOptions],
    ) -> None:
        import sklearn.base
        import sklearn.pipeline

        assert isinstance(model, sklearn.base.BaseEstimator) or isinstance(model, sklearn.pipeline.Pipeline)

        if not is_sub_model:
            target_methods = model_meta_api._get_target_methods(
                model=model,
                target_methods=kwargs.pop("target_methods", None),
                default_target_methods=_SKLModelHandler.DEFAULT_TARGET_METHODS,
            )

            def get_prediction(
                target_method_name: str, sample_input: model_types.SupportedLocalDataType
            ) -> model_types.SupportedLocalDataType:
                if not isinstance(sample_input, (pd.DataFrame, np.ndarray)):
                    sample_input = model_signature._convert_local_data_to_df(sample_input)

                target_method = getattr(model, target_method_name, None)
                assert callable(target_method)
                predictions_df = target_method(sample_input)
                return predictions_df

            model_meta = model_meta_api._validate_signature(
                model=model,
                model_meta=model_meta,
                target_methods=target_methods,
                sample_input=sample_input,
                get_prediction_fn=get_prediction,
            )

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        os.makedirs(model_blob_path, exist_ok=True)
        with open(os.path.join(model_blob_path, _SKLModelHandler.MODEL_BLOB_FILE), "wb") as f:
            cloudpickle.dump(model, f)
        base_meta = model_meta_api._ModelBlobMetadata(
            name=name, model_type=_SKLModelHandler.handler_type, path=_SKLModelHandler.MODEL_BLOB_FILE
        )
        model_meta.models[name] = base_meta
        model_meta._include_if_absent([model_meta_api.Dependency(conda_name="scikit-learn", pip_name="scikit-learn")])

    @staticmethod
    def _load_model(
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"]:
        model_blob_path = os.path.join(model_blobs_dir_path, name)
        if not hasattr(model_meta, "models"):
            raise ValueError("Ill model metadata found.")
        model_blobs_metadata = model_meta.models
        if name not in model_blobs_metadata:
            raise ValueError(f"Blob of model {name} does not exist.")
        model_blob_metadata = model_blobs_metadata[name]
        model_blob_filename = model_blob_metadata.path
        with open(os.path.join(model_blob_path, model_blob_filename), "rb") as f:
            m = cloudpickle.load(f)

        import sklearn.base
        import sklearn.pipeline

        assert isinstance(m, sklearn.base.BaseEstimator) or isinstance(m, sklearn.pipeline.Pipeline)
        return m

    @staticmethod
    def _load_as_custom_model(
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> custom_model.CustomModel:
        """Create a custom model class wrap for unified interface when being deployed. The predict method will be
        re-targeted based on target_method metadata.

        Args:
            name: Name of the model.
            model_meta: The model metadata.
            model_blobs_dir_path: Directory path to the whole model.
            kwargs: Options when loading the model.

        Returns:
            The model object as a custom model.
        """
        from snowflake.ml.model import custom_model

        def _create_custom_model(
            raw_model: Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"],
            model_meta: model_meta_api.ModelMetadata,
        ) -> Type[custom_model.CustomModel]:
            def fn_factory(
                raw_model: Union["sklearn.base.BaseEstimator", "sklearn.pipeline.Pipeline"],
                signature: model_signature.ModelSignature,
                target_method: str,
            ) -> Callable[[custom_model.CustomModel, pd.DataFrame], pd.DataFrame]:
                @custom_model.inference_api
                def fn(self: custom_model.CustomModel, X: pd.DataFrame) -> pd.DataFrame:
                    res = getattr(raw_model, target_method)(X)

                    if isinstance(res, list) and len(res) > 0 and isinstance(res[0], np.ndarray):
                        # In case of multi-output estimators, predict_proba(), decision_function(), etc., functions
                        # return a list of ndarrays. We need to deal them seperately
                        df = numpy_handler.SeqOfNumpyArrayHandler.convert_to_df(res)
                    else:
                        df = pd.DataFrame(res)

                    return model_signature_utils.rename_pandas_df(df, signature.outputs)

                return fn

            type_method_dict = {}
            for target_method_name, sig in model_meta.signatures.items():
                type_method_dict[target_method_name] = fn_factory(raw_model, sig, target_method_name)

            _SKLModel = type(
                "_SKLModel",
                (custom_model.CustomModel,),
                type_method_dict,
            )

            return _SKLModel

        raw_model = _SKLModelHandler._load_model(name, model_meta, model_blobs_dir_path, **kwargs)
        _SKLModel = _create_custom_model(raw_model, model_meta)
        skl_model = _SKLModel(custom_model.ModelContext())

        return skl_model



import functools
import importlib
import os
import pkgutil
from types import ModuleType
from typing import Any, Callable, Dict, Optional, Type, TypeVar, cast

from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._handlers import _base

_HANDLERS_BASE = "_handlers"
_MODEL_HANDLER_REGISTRY: Dict[str, Type[_base._ModelHandler[model_types.SupportedModelType]]] = dict()
_IS_HANDLER_LOADED = False


def _register_handlers() -> None:
    """
    Scan all Python modules in _HANDLERS_BASE directory and register every found non-base ModelHandler
    automatically.
    """
    model_module = importlib.import_module("snowflake.ml.model")
    model_path = model_module.__path__

    for _, name, _ in pkgutil.iter_modules(
        map(lambda x: os.path.join(x, _HANDLERS_BASE), model_path), "snowflake.ml.model._handlers."
    ):
        handler_module = importlib.import_module(name)
        if type(handler_module) == ModuleType:
            for c in dir(handler_module):
                k_class = getattr(handler_module, c)
                if (
                    isinstance(k_class, type)
                    and k_class is not _base._ModelHandler
                    and issubclass(k_class, _base._ModelHandler)
                ):
                    _MODEL_HANDLER_REGISTRY[k_class.handler_type] = k_class


F = TypeVar("F", bound=Callable[..., Any])


def ensure_handlers_registration(fn: F) -> F:
    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        global _IS_HANDLER_LOADED
        if not _IS_HANDLER_LOADED:
            _register_handlers()
            _IS_HANDLER_LOADED = True

        return fn(*args, **kwargs)

    return cast(F, wrapper)


@ensure_handlers_registration
def _find_handler(
    model: model_types.SupportedModelType,
) -> Optional[Type[_base._ModelHandler[model_types.SupportedModelType]]]:
    for handler in _MODEL_HANDLER_REGISTRY.values():
        if handler.can_handle(model):
            return handler
    return None


@ensure_handlers_registration
def _load_handler(target_model_type: str) -> Optional[Type[_base._ModelHandler[model_types.SupportedModelType]]]:
    for model_type, handler in _MODEL_HANDLER_REGISTRY.items():
        if target_model_type == model_type:
            return handler
    return None


@ensure_handlers_registration
def is_auto_signature_model(model: model_types.SupportedModelType) -> bool:
    for handler in _MODEL_HANDLER_REGISTRY.values():
        if handler.can_handle(model):
            return handler.is_auto_signature
    return False



print("import: 'snowflake.ml'")
import snowflake.ml

print("import: 'snowflake.ml.fileset'")
import snowflake.ml.fileset

print("import: 'snowflake.ml.model'")
import snowflake.ml.model

print("import: 'snowflake.ml.modeling'")
import snowflake.ml.modeling

print("import: 'snowflake.ml.registry'")
import snowflake.ml.registry

print("import: 'snowflake.ml.training_dataset'")
import snowflake.ml.training_dataset

print("import: 'snowflake.ml.utils'")
import snowflake.ml.utils




from __future__ import annotations  # for return self methods

from functools import partial
from typing import Any, Callable, Optional

from snowflake import connector, snowpark
from snowflake.ml._internal.utils import formatting


def _query_log(sql: str | None) -> str:
    """Returns the query string to log if valid. Returns empty otherwise."""
    return f" Query: {sql}" if sql else ""


def result_dimension_matcher(
    expected_rows: Optional[int], expected_cols: Optional[int], result: list[snowpark.Row], sql: str | None = None
) -> bool:
    """Check result dimensions of the collected result dataframe of a Snowflake SQL operation.

    If rows or cols are None, they are not checked.

    Args:
        expected_rows (int): Number of expected rows in the result.
        expected_cols (int): Number of expected columns in the result:
        result (List[snowpark.Row]): Collected dataframe containing the result status of a Snowflake SQL operation.
        sql (str): Query string to be included in the error messages. Will not affect the functionality otherwise.

    Returns:
        true if `expected_{rows|cols}` number of rows and columns found in `result`.

    Raises:
        DataError: In case the validation failed.
    """
    actual_rows = len(result)
    if expected_rows is not None and actual_rows != expected_rows:
        raise connector.DataError(
            formatting.unwrap(
                f"""Query Result did not match expected number of rows. Expected {expected_rows} rows, found:
                {actual_rows} rows. Result from operation was: {result}.{_query_log(sql)}"""
            )
        )

    if expected_cols is not None:
        if not result:
            raise connector.DataError(
                formatting.unwrap(
                    f"""Query Result is empty but a number of columns expected. Expected {expected_cols} columns.
                    Result from operation was: {result}.{_query_log(sql)}"""
                )
            )

        actual_cols = len(result[0])
        if actual_cols != expected_cols:
            raise connector.DataError(
                formatting.unwrap(
                    f"""Query Result did not match expected number of columns. Expected {expected_cols} columns,
                    found {actual_cols} columns. Result from operation was: {result}.{_query_log(sql)}"""
                )
            )
    return True


def column_name_matcher(expected_col_name: str, result: list[snowpark.Row], sql: str | None = None) -> bool:
    """Returns true if `expected_col_name` is found. Raise exception otherwise."""
    if not result:
        raise connector.DataError(f"Query Result is empty.{_query_log(sql)}")
    if expected_col_name not in result[0]:
        raise connector.DataError(
            formatting.unwrap(
                f"""Query Result did not have expected column {expected_col_name}. Result from operation was:
                    {result}.{_query_log(sql)}"""
            )
        )
    return True


def cell_value_by_column_matcher(
    row_idx: int, expected_col_name: str, expected_value: Any, result: list[snowpark.Row], sql: str | None = None
) -> bool:
    """Returns true if `expected_col_name` is found in `result` and the value in row `row_idx` matches `value`. Raises a
    connector.DataError otherwise.

    Args:
        row_idx (int): index of the row to check the value
        expected_col_name (str): name of the column to check
        expected_value (Any): What to compare the value in the result against.
        result (list[snowpark.Row]): result to check. Typically the result of a .collect() call on a DataFrame.
        sql (Optional[str]): Sql string that generated this result. This is optional and for more detailed error
            messages only.

    Returns:
        True iff the given result matches the expectation given by the arguments.

    Raises:
        DataError: There is a mismatch between the given result and the stated expectation.
    """
    if len(result) <= row_idx or len(result[row_idx]) < 1:
        raise connector.DataError(
            formatting.unwrap(
                f"""Query Result did not have required number of rows x col [{row_idx}][{expected_col_name}]. Result
                from operation was: {result}.{_query_log(sql)}"""
            )
        )
    if expected_col_name not in result[row_idx]:
        raise connector.DataError(
            formatting.unwrap(
                f"""Query Result did not have the expected value column '{expected_col_name}' in row {row_idx}.
                Actual row looked like {result[row_idx]}.{_query_log(sql)}"""
            )
        )
    if result[row_idx][expected_col_name] != expected_value:
        raise connector.DataError(
            formatting.unwrap(
                f"""Query Result did not have the expected value '{expected_value}' at expected position
                [{row_idx}][{expected_col_name}]. Actual value at position [{row_idx}][{expected_col_name}] was
                '{result[row_idx][expected_col_name]}'.{_query_log(sql)}"""
            )
        )
    return True


_DEFAULT_MATCHERS = [
    partial(result_dimension_matcher, 1, 1),
    partial(column_name_matcher, "status"),
]


class ResultValidator:
    """Convenience wrapper for validation SnowPark DataFrames that are returned as the result of session operations.

    Usage Example:
        result = (
            ResultValidator(
                session=self._session,
                query="UPDATE table SET NAME = 'name'",
            )
            .has_dimensions(expected_rows=1, expected_cols=1)
            .has_partial_match(row_idx=0, col_idx=0, expected_value="number of rows updated=1")
            .validate()
        )
    """

    def __init__(self, result: list[snowpark.Row], query: str | None = None) -> None:
        self._result: list[snowpark.Row] = result
        self._query: str | None = query
        self._success_matchers: list[Callable[[list[snowpark.Row], str], bool]] = []

    def has_dimensions(self, expected_rows: int | None = None, expected_cols: int | None = None) -> ResultValidator:
        """Validate that the result of the operation has the right shape of `expected_rows` rows and `expected_cols`
        columns.

        Args:
            expected_rows: Number of rows expected in the result.
            expected_cols: Number of columns expected in the result.

        Returns:
            ResultValidator object (self)
        """
        self._success_matchers.append(partial(result_dimension_matcher, expected_rows, expected_cols))
        return self

    def has_column(self, expected_col_name: str) -> ResultValidator:
        """Validate that the a column with the name `expected_column_name` exists in the result.

        Args:
            expected_col_name: Name of the column that is expected to be present in the result (case sensitive).

        Returns:
            ResultValidator object (self)
        """
        self._success_matchers.append(partial(column_name_matcher, expected_col_name))
        return self

    def has_named_value_match(self, row_idx: int, col_name: str, expected_value: Any) -> ResultValidator:
        """Validate that the column `col_name` in row `row_idx` of ther results exists and matches `expected_value`.

        Args:
            row_idx: Row index of the cell that needs to match.
            col_name: Column name of the cell that needs to match.
            expected_value: Value that the cell needs to match. For strings it is treated as a substring match, all
                other types will expect an exact match.

        Returns:
            ResultValidator object (self)
        """
        self._success_matchers.append(partial(cell_value_by_column_matcher, row_idx, col_name, expected_value))
        return self

    def insertion_success(self, expected_num_rows: int) -> ResultValidator:
        """Validate that `expected_num_rows` have been inserted successfully.

        Args:
            expected_num_rows: Number of rows that are expected to be inserted successfully.

        Returns:
            ResultValidator object (self)
        """
        self._success_matchers.append(
            partial(cell_value_by_column_matcher, 0, "number of rows inserted", expected_num_rows)
        )
        return self

    def deletion_success(self, expected_num_rows: int) -> ResultValidator:
        """Validate that `expected_num_rows` have been deleted successfully.

        Args:
            expected_num_rows: Number of rows that are expected to be deleted successfully.

        Returns:
            ResultValidator object (self)
        """
        self._success_matchers.append(
            partial(cell_value_by_column_matcher, 0, "number of rows deleted", expected_num_rows)
        )
        return self

    def _get_result(self) -> list[snowpark.Row]:
        """Return the given result DataFrame."""
        return self._result

    def validate(self) -> list[snowpark.Row]:
        """Execute the query and validate the result.

        Returns:
            Query result.
        """
        if len(self._success_matchers) == 0:
            self._success_matchers = _DEFAULT_MATCHERS  # type: ignore
        result = self._get_result()
        for matcher in self._success_matchers:
            assert matcher(result, self._query)  # type: ignore
        return result


class SqlResultValidator(ResultValidator):
    """Convenience wrapper for validation of SQL queries.

    Usage Example:
        result = (
            SqlResultValidator(
                session=self._session,
                query="UPDATE table SET NAME = 'name'",
            )
            .has_dimensions(expected_rows=1, expected_cols=1)
            .has_partial_match(row_idx=0, col_idx=0, expected_value="number of rows updated=1")
            .validate()
        )
    """

    def __init__(self, session: snowpark.Session, query: str) -> None:
        self._session: snowpark.Session = session
        self._query: str = query
        self._success_matchers: list[Callable[[list[snowpark.Row], str], bool]] = []

    def _get_result(self) -> list[snowpark.Row]:
        """Collect the result of the given SQL query."""
        return self._session.sql(self._query).collect()



"""String formatting utilities for general use in the SnowML Reposiory.

This file contains a collection of utilities that help with formatting strings. Functionality is not limited to tests
only. Anything that is re-usable across different modules and related to string formatting should go here.
"""

import re
from dataclasses import dataclass
from datetime import datetime
from typing import Any

_WHITESPACE_COMPACT_RE = re.compile(r"\s+")
_WHITESPACE_NO_NEWLINE_RE = re.compile(r"[ \t\r\f\v]+")
_WHITESPACE_COMPACT_KEEP_NEWLINE_RE = re.compile(r"([ \t\r\f\v]*\n[ \t\r\f\v]*)")


@dataclass
class SqlStr:
    """Class to represent SQL strings for formatting.

    This is mainly to distinguish SQL from regular strings."""

    sql: str

    def __init__(self, sql: str) -> None:
        self.sql = sql

    def __repr__(self) -> str:
        return self.sql


def format_value_for_select(value: Any) -> str:
    """Format a value for inclusion in a SELECT query. returns a string with the correct formatting.

    Currently supported types:
        str: Enclose in single quotes and escape single quotes inside the string.
        datetime: Convert isoformat string and instruct the backend to interpret the string as a timestamp.
        dict: Convert keys to strings and recursively handle values. Instruct the backend to interpret the list of keys
            and values as an object.

        Everything else we attempt to convert to string but will not enclose in quotes.

    Args:
        value: Value to be formatted.

    Returns:
        String with the formatted value.
    """

    # Have to use an explicit comparison with None as "The truth value of an array with more than one element is
    # ambiguous." for numpy arrays.
    if value is None:
        return "null"

    if isinstance(value, str):
        return "'" + value.replace("'", "\\'") + "'"
    elif isinstance(value, SqlStr):
        return repr(value)
    elif isinstance(value, datetime):
        return "TO_TIMESTAMP('" + value.isoformat() + "')"
    elif isinstance(value, dict):
        # Converting all keys to strings and recursively format the values. When iterating over dictionaries, items
        # appear in random order due to the hashing of the keys. We ensure reproducibility by sorting the items during
        # formatting.

        return (
            "OBJECT_CONSTRUCT("
            + ",".join([f"'{k}',{format_value_for_select(v)}" for k, v in sorted(value.items())])
            + ")"
        )
    elif hasattr(value, "__iter__"):
        # If value is iterable (e.g. list or array) convert it to an array recursively.
        return "ARRAY_CONSTRUCT(" + ",".join([format_value_for_select(x) for x in value]) + ")"
    else:
        return str(value)


def unwrap(text: str, keep_newlines: bool = False) -> str:
    """Unwraps a string into a single line while preserving word boundaries. Leading and trailing spaces are removed.

    Args:
        text: Text to unwrap.
        keep_newlines: Keep newlines when formatting.

    Returns:
        Unwrapped text.
    """

    # Normalize whitespace:
    if keep_newlines:
        # Three stages:
        # 1. (innermost) Normalize all non-newline whitespace to spaces.
        # 2. (middle)    Collapse any sequence of newline plus any non-newline whitespace into just newline.
        # 3. (outter)    Strip leading and trailing whitespace.
        return _WHITESPACE_COMPACT_KEEP_NEWLINE_RE.sub("\n", _WHITESPACE_NO_NEWLINE_RE.sub(" ", text)).strip()
    else:
        # Normalize all whitespace into single spaces and strip leading and trailing whitespace.
        return _WHITESPACE_COMPACT_RE.sub(" ", text).strip()



#!/usr/bin/env python3
#
# Copyright (c) 2012-2022 Snowflake Computing Inc. All rights reserved.
#
from typing import Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.preprocessing import _data as sklearn_preprocessing_data

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml.modeling.framework import _utils, base


class StandardScaler(base.BaseTransformer):
    r"""
    Standardizes features by removing the mean and scaling to unit variance. Values must be of float type.

    For more details on what this transformer does, see [sklearn.preprocessing.StandardScaler]
    (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html).

    Args:
        with_mean: If True, center the data before scaling.
        with_std: If True, scale the data unit variance (i.e. unit standard deviation).
        input_cols: The name(s) of one or more columns in a DataFrame containing a feature to be scaled.
        output_cols: The name(s) of one or more columns in a DataFrame in which results will be stored. The number of
            columns specified must match the number of input columns.
        drop_input_cols: Remove input columns from output if set True. False by default.

    Attributes:
        scale_: Dictionary mapping input column names to relative scaling factor to achieve zero mean and unit variance.
            If a variance is zero, unit variance could not be achieved, and the data is left as-is, giving a scaling
            factor of 1. None if with_std is False.
        mean_: Dictionary mapping input column name to the mean value for that feature. None if with_mean is False.
        var_: Dictionary mapping input column name to the variance for that feature. Used to compute scale_. None if
            with_std is False
    """

    def __init__(
        self,
        *,
        with_mean: bool = True,
        with_std: bool = True,
        input_cols: Optional[Union[str, Iterable[str]]] = None,
        output_cols: Optional[Union[str, Iterable[str]]] = None,
        drop_input_cols: Optional[bool] = False,
    ) -> None:
        """
        Standardize features by removing the mean and scaling to unit variance.

        Args:
            with_mean: If True, center the data before scaling.
                This does not work (and will raise an exception) when attempted on
                sparse matrices, because centering them entails building a dense
                matrix which in common use cases is likely to be too large to fit in
                memory.
            with_std: If True, scale the data to unit variance (or equivalently,
                unit standard deviation).
            input_cols: Single or multiple input columns.
            output_cols: Single or multiple output columns.
            drop_input_cols: Remove input columns from output if set True. False by default.

        Attributes:
            scale_: dict {column_name: value} or None
                Per feature relative scaling of the data to achieve zero mean and unit
                variance. If a variance is zero, we can't achieve unit variance, and
                the data is left as-is, giving a scaling factor of 1. `scale_` is equal
                to `None` when `with_std=False`.
            mean_: dict {column_name: value} or None
                The mean value for each feature in the training set.
                Equal to ``None`` when ``with_mean=False``.
            var_: dict {column_name: value} or None
                The variance for each feature in the training set. Used to compute
                `scale_`. Equal to ``None`` when ``with_std=False``.
        """
        self.with_mean = with_mean
        self.with_std = with_std

        self.scale_: Optional[Dict[str, float]] = {} if with_std else None
        self.mean_: Optional[Dict[str, float]] = {} if with_mean else None
        self.var_: Optional[Dict[str, float]] = {} if with_std else None

        self.custom_states: List[str] = []
        if with_mean:
            self.custom_states.append(_utils.NumericStatistics.MEAN)
        if with_std:
            self.custom_states.append(_utils.NumericStatistics.VAR_POP)
            self.custom_states.append(_utils.NumericStatistics.STDDEV_POP)

        super().__init__(drop_input_cols=drop_input_cols, custom_states=self.custom_states)

        self.set_input_cols(input_cols)
        self.set_output_cols(output_cols)

    def _reset(self) -> None:
        """
        Reset internal data-dependent state of the scaler, if necessary.
        __init__ parameters are not touched.
        """
        super()._reset()
        if hasattr(self, "scale_"):
            self.scale_ = {} if self.with_std else None
        if hasattr(self, "mean_"):
            self.mean_ = {} if self.with_mean else None
        if hasattr(self, "var_"):
            self.var_ = {} if self.with_std else None

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def fit(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> "StandardScaler":
        """
        Compute mean and std values of the dataset.

        Args:
            dataset: Input dataset.

        Returns:
            Fitted scaler.
        """
        super()._check_input_cols()
        super()._check_dataset_type(dataset)
        self._reset()

        if isinstance(dataset, pd.DataFrame):
            self._fit_sklearn(dataset)
        else:
            self._fit_snowpark(dataset)

        self._is_fitted = True
        return self

    def _fit_sklearn(self, dataset: pd.DataFrame) -> None:
        dataset = self._use_input_cols_only(dataset)
        sklearn_scaler = self._create_unfitted_sklearn_object()
        sklearn_scaler.fit(dataset[self.input_cols])

        for i, input_col in enumerate(self.input_cols):
            if self.mean_ is not None:
                self.mean_[input_col] = float(sklearn_scaler.mean_[i])
            if self.scale_ is not None:
                self.scale_[input_col] = float(sklearn_scaler.scale_[i])
            if self.var_ is not None:
                self.var_[input_col] = float(sklearn_scaler.var_[i])

    def _fit_snowpark(self, dataset: snowpark.DataFrame) -> None:
        computed_states = self._compute(dataset, self.input_cols, self.custom_states)

        # assign states to the object
        for input_col in self.input_cols:
            numeric_stats = computed_states[input_col]

            if self.mean_ is not None:
                self.mean_[input_col] = float(numeric_stats[_utils.NumericStatistics.MEAN])

            if self.var_ is not None:
                self.var_[input_col] = float(numeric_stats[_utils.NumericStatistics.VAR_POP])

            if self.scale_ is not None:
                self.scale_[input_col] = sklearn_preprocessing_data._handle_zeros_in_scale(
                    float(numeric_stats[_utils.NumericStatistics.STDDEV_POP])
                )

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    @telemetry.add_stmt_params_to_df(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def transform(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> Union[snowpark.DataFrame, pd.DataFrame]:
        """
        Perform standardization by centering and scaling.

        Args:
            dataset: Input dataset.

        Returns:
            transformed_dataset: Output dataset.
        """
        self._enforce_fit()
        super()._check_input_cols()
        super()._check_output_cols()
        super()._check_dataset_type(dataset)

        if isinstance(dataset, snowpark.DataFrame):
            output_df = self._transform_snowpark(dataset)
        else:
            output_df = self._transform_sklearn(dataset)

        return self._drop_input_columns(output_df) if self._drop_input_cols is True else output_df

    def _transform_snowpark(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        """
        Perform standardization by centering and scaling on
        Snowpark dataframe.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset.
        """
        passthrough_columns = [c for c in dataset.columns if c not in self.output_cols]
        output_columns = []
        for _, input_col in enumerate(self.input_cols):
            output_column = dataset[input_col]

            if self.mean_ is not None:
                output_column = output_column - self.mean_[input_col]
            if self.scale_ is not None:
                output_column = output_column / self.scale_[input_col]

            output_columns.append(output_column)

        transformed_dataset: snowpark.DataFrame = dataset.with_columns(self.output_cols, output_columns)
        # Reorder columns. Passthrough columns are added at the right to the output of the transformers.
        transformed_dataset = transformed_dataset[self.output_cols + passthrough_columns]
        return transformed_dataset

    def _create_unfitted_sklearn_object(self) -> preprocessing.StandardScaler:
        return preprocessing.StandardScaler(with_mean=self.with_mean, with_std=self.with_std)

    def _create_sklearn_object(self) -> preprocessing.StandardScaler:
        """
        Get an equivalent sklearn StandardScaler.

        Returns:
            The Sklearn StandardScaler.
        """
        scaler = self._create_unfitted_sklearn_object()
        if self._is_fitted:
            scaler.scale_ = self._convert_attribute_dict_to_ndarray(self.scale_, np.float64)
            scaler.mean_ = self._convert_attribute_dict_to_ndarray(self.mean_, np.float64)
            scaler.var_ = self._convert_attribute_dict_to_ndarray(self.var_, np.float64)
        return scaler



import os
import shutil
import string
from abc import ABC

from packaging import version

from snowflake.ml.model import _model_meta
from snowflake.ml.model._deploy_client.utils import constants


class DockerContext(ABC):
    """
    Constructs the Docker context directory required for image building.
    """

    def __init__(self, context_dir: str, model_meta: _model_meta.ModelMetadata) -> None:
        """Initialization

        Args:
            context_dir: Path to context directory.
            model_meta: Model Metadata
        """
        self.context_dir = context_dir
        self.model_meta = model_meta

    def build(self) -> None:
        """
        Generates and/or moves resources into the Docker context directory.Rename the random model directory name to
        constant "model_dir" instead for better readability.
        """
        self._generate_inference_code()
        self._copy_entrypoint_script_to_docker_context()
        self._copy_model_env_dependency_to_docker_context()
        self._generate_docker_file()

    def _copy_entrypoint_script_to_docker_context(self) -> None:
        """Copy gunicorn_run.sh entrypoint to docker context directory."""
        path = os.path.join(os.path.dirname(__file__), constants.ENTRYPOINT_SCRIPT)
        assert os.path.exists(path), f"Run script file missing at path: {path}"
        shutil.copy(path, os.path.join(self.context_dir, constants.ENTRYPOINT_SCRIPT))

    def _copy_model_env_dependency_to_docker_context(self) -> None:
        """
        Convert model dependencies to files from model metadata.
        """
        self.model_meta.save_model_metadata(self.context_dir)

    def _generate_docker_file(self) -> None:
        """
        Generates dockerfile based on dockerfile template.
        """
        docker_file_path = os.path.join(self.context_dir, "Dockerfile")
        docker_file_template = os.path.join(os.path.dirname(__file__), "templates/dockerfile_template")

        if self.model_meta.cuda_version:
            cuda_version_parsed = version.parse(self.model_meta.cuda_version)
            cuda_version_str = f"{cuda_version_parsed.major}.{cuda_version_parsed.minor}"
        else:
            cuda_version_str = ""

        with open(docker_file_path, "w", encoding="utf-8") as dockerfile, open(
            docker_file_template, encoding="utf-8"
        ) as template:
            dockerfile_content = string.Template(template.read()).safe_substitute(
                {
                    # TODO(shchen): SNOW-835411, Support overwriting base image
                    "base_image": "mambaorg/micromamba:1.4.3",
                    "model_env_folder": constants.MODEL_ENV_FOLDER,
                    "inference_server_dir": constants.INFERENCE_SERVER_DIR,
                    "entrypoint_script": constants.ENTRYPOINT_SCRIPT,
                    # Instead of omitting this ENV var when no CUDA required, we explicitly set it to empty to override
                    # as no CUDA is detected thus it won't be affected by the existence of CUDA in base image.
                    # https://conda.io/projects/conda/en/latest/user-guide/tasks/manage-virtual.html
                    "cuda_override_env": cuda_version_str,
                }
            )
            dockerfile.write(dockerfile_content)

    def _generate_inference_code(self) -> None:
        """
        Generates inference code based on the app template and creates a folder named 'server' to house the inference
        server code.
        """
        inference_server_folder_path = os.path.join(os.path.dirname(__file__), constants.INFERENCE_SERVER_DIR)
        destination_folder_path = os.path.join(self.context_dir, constants.INFERENCE_SERVER_DIR)
        ignore_patterns = shutil.ignore_patterns("BUILD.bazel", "*test.py", "*.\\.*", "__pycache__")
        shutil.copytree(inference_server_folder_path, destination_folder_path, ignore=ignore_patterns)



from collections import abc
from typing import TYPE_CHECKING, List, Literal, Optional, Sequence, Union

import numpy as np
import pandas as pd
from typing_extensions import TypeGuard

from snowflake.ml._internal import type_utils
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._signatures import base_handler, core

if TYPE_CHECKING:
    import tensorflow


class SeqOfTensorflowTensorHandler(
    base_handler.BaseDataHandler[Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]]
):
    @staticmethod
    def can_handle(
        data: model_types.SupportedDataType,
    ) -> TypeGuard[Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]]:
        if not isinstance(data, abc.Sequence):
            return False
        if len(data) == 0:
            return False
        if type_utils.LazyType("tensorflow.Tensor").isinstance(data[0]) or type_utils.LazyType(
            "tensorflow.Variable"
        ).isinstance(data[0]):
            return all(
                type_utils.LazyType("tensorflow.Tensor").isinstance(data_col)
                or type_utils.LazyType("tensorflow.Variable").isinstance(data_col)
                for data_col in data
            )
        return False

    @staticmethod
    def count(data: Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]) -> int:
        import tensorflow as tf

        rows = []
        for data_col in data:
            shapes = data_col.shape.as_list()
            if data_col.shape == tf.TensorShape(None) or (not shapes) or (shapes[0] is None):
                # Unknown shape array
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError("Data Validation Error: Unknown shape data is found."),
                )
            # Make mypy happy
            assert isinstance(shapes[0], int)

            rows.append(shapes[0])

        return min(rows)

    @staticmethod
    def truncate(
        data: Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]
    ) -> Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]:
        return [
            data_col[
                : min(SeqOfTensorflowTensorHandler.count(data), SeqOfTensorflowTensorHandler.SIG_INFER_ROWS_COUNT_LIMIT)
            ]
            for data_col in data
        ]

    @staticmethod
    def validate(data: Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]) -> None:
        import tensorflow as tf

        for data_col in data:
            if data_col.shape == tf.TensorShape(None) or any(dim is None for dim in data_col.shape.as_list()):
                # Unknown shape array
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError("Data Validation Error: Unknown shape data is found."),
                )

            if data_col.shape == tf.TensorShape([0]):
                # Empty array
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError("Data Validation Error: Empty data is found."),
                )

            if data_col.shape == tf.TensorShape([1]) or data_col.shape == tf.TensorShape([]):
                # scalar
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError("Data Validation Error: Scalar data is found."),
                )

    @staticmethod
    def infer_signature(
        data: Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]], role: Literal["input", "output"]
    ) -> Sequence[core.BaseFeatureSpec]:
        feature_prefix = f"{SeqOfTensorflowTensorHandler.FEATURE_PREFIX}_"
        features: List[core.BaseFeatureSpec] = []
        role_prefix = (
            SeqOfTensorflowTensorHandler.INPUT_PREFIX if role == "input" else SeqOfTensorflowTensorHandler.OUTPUT_PREFIX
        ) + "_"

        for i, data_col in enumerate(data):
            dtype = core.DataType.from_numpy_type(data_col.dtype.as_numpy_dtype)
            ft_name = f"{role_prefix}{feature_prefix}{i}"
            if len(data_col.shape) == 1:
                features.append(core.FeatureSpec(dtype=dtype, name=ft_name))
            else:
                ft_shape = tuple(data_col.shape[1:])
                features.append(core.FeatureSpec(dtype=dtype, name=ft_name, shape=ft_shape))
        return features

    @staticmethod
    def convert_to_df(
        data: Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]], ensure_serializable: bool = True
    ) -> pd.DataFrame:
        if ensure_serializable:
            return pd.DataFrame({i: data_col.numpy().tolist() for i, data_col in enumerate(iterable=data)})
        return pd.DataFrame({i: list(data_col.numpy()) for i, data_col in enumerate(iterable=data)})

    @staticmethod
    def convert_from_df(
        df: pd.DataFrame, features: Optional[Sequence[core.BaseFeatureSpec]] = None
    ) -> Sequence[Union["tensorflow.Tensor", "tensorflow.Variable"]]:
        import tensorflow as tf

        res = []
        if features:
            for feature in features:
                if isinstance(feature, core.FeatureGroupSpec):
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.NOT_IMPLEMENTED,
                        original_exception=NotImplementedError("FeatureGroupSpec is not supported."),
                    )
                assert isinstance(feature, core.FeatureSpec), "Invalid feature kind."
                res.append(
                    tf.convert_to_tensor(np.stack(df[feature.name].to_numpy()).astype(feature._dtype._numpy_type))
                )
            return res
        return [tf.convert_to_tensor(np.stack(df[col].to_numpy())) for col in df]



from typing import Any, Dict, Iterator

import fsspec
import numpy.typing as npt
from torchdata.datapipes.iter import IterDataPipe

from snowflake.ml.fileset import parquet_parser


class ReadAndParseParquet(IterDataPipe):
    """Read and parse the parquet files yield batched numpy array in dict.

    Args:
        input_datapipe: A datapipe of input parquet file URIs to read and parse.
            Note that the datapipe must be finite.
        filesystem: A fsspec/pyarrow file system that is used to open given file URIs.
        batch_size: Specifies the size of each batch that will be yield
        shuffle: Whether the data in the file will be shuffled. If set to be true, it will first randomly shuffle
            the order of files, and then shuflle the order of rows in each file.
        drop_last_batch: Whehter the last batch of data should be dropped. If set to be true, then the last batch will
            get dropped if its size is smaller than the given batch_size.

    Returns:
        A PyTorch iterable datapipe that yields batched numpy array in dict. The keys will be the column names in
        the parquet files.

    Example:
        >>> from snowflake.ml.fileset import sfcfs, torch_datapipe
        >>> from torchdata.datapipes.iter import FSSpecFileLister
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> fs = sfcfs.SFFileSystem(conn)
        >>> filedp = FSSpecFileLister(root=dir_path, masks="*.parquet", mode="rb", sf_connection=conn)
        >>> parquet_dp = torch_datapipe.ReadAndParseParquet(file_dp, fs, batch_size = 2)
        >>> for batch in parquet_dp:
        >>>     print(batch)
    ----
    {'_COL_1': [32.5000, 6.0000], '_COL_2': [-73.9542, -73.9875]}
    """

    def __init__(
        self,
        input_datapipe: IterDataPipe[str],
        filesystem: fsspec.AbstractFileSystem,
        batch_size: int,
        shuffle: bool,
        drop_last_batch: bool,
    ) -> None:
        self._input_datapipe = input_datapipe
        self._fs = filesystem
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._drop_last_batch = drop_last_batch

    def __iter__(self) -> Iterator[Dict[str, npt.NDArray[Any]]]:
        yield from parquet_parser.ParquetParser(
            list(self._input_datapipe), self._fs, self._batch_size, self._shuffle, self._drop_last_batch
        )



import warnings
from typing import Any, Dict, Optional


class HuggingFacePipelineModel:
    def __init__(
        self,
        task: str = None,
        model: Optional[str] = None,
        *,
        revision: Optional[str] = None,
        token: Optional[str] = None,
        trust_remote_code: Optional[bool] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Utility factory method to build a wrapper over transformers [`Pipeline`].
        When deploying, this wrapper will create a real pipeline object and loading tokenizers and models.

        For pipelines docs, please refer:
        https://huggingface.co/docs/transformers/en/main_classes/pipelines#transformers.pipeline

        Args:
            task: The task that pipeline will be used. If None it would be inferred from model.
                For available tasks, please refer Transformers's documentation. Defaults to None.
            model: The model that will be used by the pipeline to make predictions. This can only be a model identifier
                currently. If not provided, the default for the `task` will be loaded. Defaults to None.
            revision: When passing a task name or a string model identifier: The specific model version to use. It can
                be a branch name, a tag name, or a commit id, since we use a git-based system for storing models and
                other artifacts on huggingface.co, so `revision` can be any identifier allowed by git. Defaults to None.
            token: The token to use as HTTP bearer authorization for remote files. Defaults to None.
            trust_remote_code: Whether or not to allow for custom code defined on the Hub in their own modeling,
                configuration, tokenization or even pipeline files. This option should only be set to `True` for
                repositories you trust and in which you have read the code, as it will execute code present on the Hub.
                Defaults to None.
            model_kwargs: Additional dictionary of keyword arguments passed along to the model's `from_pretrained(...,`.
                Defaults to None.
            kwargs: Additional keyword arguments passed along to the specific pipeline init (see the documentation for
                the corresponding pipeline class for possible values).

        Return:
            A wrapper over transformers [`Pipeline`].

        Raises:
            RuntimeError: Raised when the input argument cannot determine the pipeline.
            ValueError: Raised when the pipeline contains remote code but trust_remote_code is not set or False.
            ValueError: Raised when having conflicting arguments.
        """
        import transformers

        config = kwargs.get("config", None)
        tokenizer = kwargs.get("tokenizer", None)
        framework = kwargs.get("framework", None)
        feature_extractor = kwargs.get("feature_extractor", None)

        # ==== Start pipeline logic from transformers ====
        if model_kwargs is None:
            model_kwargs = {}

        use_auth_token = model_kwargs.pop("use_auth_token", None)
        if use_auth_token is not None:
            warnings.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers.", FutureWarning
            )
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )
            token = use_auth_token

        hub_kwargs = {
            "revision": revision,
            "token": token,
            "trust_remote_code": trust_remote_code,
            "_commit_hash": None,
        }

        if task is None and model is None:
            raise RuntimeError(
                "Impossible to instantiate a pipeline without either a task or a model being specified. "
            )

        if model is None and tokenizer is not None:
            raise RuntimeError(
                "Impossible to instantiate a pipeline with tokenizer specified but not the model as the provided"
                " tokenizer may not be compatible with the default model. Please provide an identifier to a pretrained"
                " model when providing tokenizer."
            )
        if model is None and feature_extractor is not None:
            raise RuntimeError(
                "Impossible to instantiate a pipeline with feature_extractor specified but not the model as the"
                "provided feature_extractor may not be compatible with the default model. Please provide an identifier"
                " to a pretrained model when providing feature_extractor."
            )

        # ==== End pipeline logic from transformers ====

        # We only support string as model argument.

        if model is not None and not isinstance(model, str):
            raise RuntimeError(
                "Impossible to use non-string model as input for HuggingFacePipelineModel. Use transformers.Pipeline"
                " object if required."
            )

        # ==== Start pipeline logic (Config) from transformers ====

        # Config is the primordial information item.
        # Instantiate config if needed
        if isinstance(config, str):
            config_obj = transformers.AutoConfig.from_pretrained(
                config, _from_pipeline=task, **hub_kwargs, **model_kwargs
            )
            hub_kwargs["_commit_hash"] = config_obj._commit_hash
        elif config is None:
            config_obj = transformers.AutoConfig.from_pretrained(
                model, _from_pipeline=task, **hub_kwargs, **model_kwargs
            )
            hub_kwargs["_commit_hash"] = config_obj._commit_hash
        # We only support string as config argument.
        else:
            raise RuntimeError(
                "Impossible to use non-string config as input for HuggingFacePipelineModel. Use transformers.Pipeline"
                " object if required."
            )

        # ==== Start pipeline logic (Task) from transformers ====

        custom_tasks = {}
        if config_obj is not None and len(getattr(config_obj, "custom_pipelines", {})) > 0:
            custom_tasks = config_obj.custom_pipelines
            if task is None and trust_remote_code is not False:
                if len(custom_tasks) == 1:
                    task = list(custom_tasks.keys())[0]
                else:
                    raise RuntimeError(
                        "We can't infer the task automatically for this model as there are multiple tasks available. "
                        f"Pick one in {', '.join(custom_tasks.keys())}"
                    )

        if task is None and model is not None:
            if not isinstance(model, str):
                raise RuntimeError(
                    "Inferring the task automatically requires to check the hub with a model_id defined as a `str`."
                    f"{model} is not a valid model_id."
                )
            task = transformers.pipelines.get_task(model, token)

        # Retrieve the task
        if task in custom_tasks:
            normalized_task = task
            targeted_task, task_options = transformers.pipelines.clean_custom_task(custom_tasks[task])
            if not trust_remote_code:
                raise ValueError(
                    "Loading this pipeline requires you to execute the code in the pipeline file in that"
                    " repo on your local machine. Make sure you have read the code there to avoid malicious use, then"
                    " set the option `trust_remote_code=True` to remove this error."
                )
        else:
            normalized_task, targeted_task, task_options = transformers.pipelines.check_task(task)

        # ==== Start pipeline logic (Model) from transformers ====

        # Use default model/config/tokenizer for the task if no model is provided
        if model is None:
            # At that point framework might still be undetermined
            model, default_revision = transformers.pipelines.get_default_model_and_revision(
                targeted_task, framework, task_options
            )
            revision = revision if revision is not None else default_revision
            warnings.warn(
                f"No model was supplied, defaulted to {model} and revision"
                f" {revision} ({transformers.pipelines.HUGGINGFACE_CO_RESOLVE_ENDPOINT}/{model}).\n"
                "Using a pipeline without specifying a model name and revision in production is not recommended."
            )
            if config is None and isinstance(model, str):
                config_obj = transformers.AutoConfig.from_pretrained(
                    model, _from_pipeline=task, **hub_kwargs, **model_kwargs
                )
                hub_kwargs["_commit_hash"] = config_obj._commit_hash

        if kwargs.get("device_map", None) is not None:
            if "device_map" in model_kwargs:
                raise ValueError(
                    'You cannot use both `pipeline(... device_map=..., model_kwargs={"device_map":...})` as those'
                    " arguments might conflict, use only one.)"
                )
            if kwargs.get("device", None) is not None:
                warnings.warn(
                    "Both `device` and `device_map` are specified. `device` will override `device_map`. You"
                    " will most likely encounter unexpected behavior. Please remove `device` and keep `device_map`."
                )

        # ==== End pipeline logic from transformers ====

        self.task = normalized_task
        self.model = model
        self.revision = revision
        self.token = token
        self.trust_remote_code = trust_remote_code
        self.model_kwargs = model_kwargs
        self.__dict__.update(kwargs)



# mypy: disable-error-code="import"
import os
from typing import TYPE_CHECKING, Callable, Optional, Type, Union

import numpy as np
import pandas as pd
from typing_extensions import TypeGuard, Unpack

from snowflake.ml._internal import type_utils
from snowflake.ml.model import (
    _model_meta as model_meta_api,
    custom_model,
    model_signature,
    type_hints as model_types,
)
from snowflake.ml.model._handlers import _base
from snowflake.ml.model._signatures import numpy_handler, utils as model_signature_utils

if TYPE_CHECKING:
    import xgboost


class _XGBModelHandler(_base._ModelHandler[Union["xgboost.Booster", "xgboost.XGBModel"]]):
    """Handler for XGBoost based model.

    Currently xgboost.XGBModel based classes are supported.
    """

    handler_type = "xgboost"
    MODEL_BLOB_FILE = "model.ubj"
    DEFAULT_TARGET_METHODS = ["apply", "predict", "predict_proba"]

    @staticmethod
    def can_handle(model: model_types.SupportedModelType) -> TypeGuard[Union["xgboost.Booster", "xgboost.XGBModel"]]:
        return (
            type_utils.LazyType("xgboost.Booster").isinstance(model)
            or type_utils.LazyType("xgboost.XGBModel").isinstance(model)
        ) and any(
            (hasattr(model, method) and callable(getattr(model, method, None)))
            for method in _XGBModelHandler.DEFAULT_TARGET_METHODS
        )

    @staticmethod
    def cast_model(
        model: model_types.SupportedModelType,
    ) -> Union["xgboost.Booster", "xgboost.XGBModel"]:
        import xgboost

        assert isinstance(model, xgboost.Booster) or isinstance(model, xgboost.XGBModel)

        return model

    @staticmethod
    def _save_model(
        name: str,
        model: Union["xgboost.Booster", "xgboost.XGBModel"],
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        sample_input: Optional[model_types.SupportedDataType] = None,
        is_sub_model: Optional[bool] = False,
        **kwargs: Unpack[model_types.XGBModelSaveOptions],
    ) -> None:
        import xgboost

        assert isinstance(model, xgboost.Booster) or isinstance(model, xgboost.XGBModel)

        if not is_sub_model:
            target_methods = model_meta_api._get_target_methods(
                model=model,
                target_methods=kwargs.pop("target_methods", None),
                default_target_methods=_XGBModelHandler.DEFAULT_TARGET_METHODS,
            )

            def get_prediction(
                target_method_name: str, sample_input: model_types.SupportedLocalDataType
            ) -> model_types.SupportedLocalDataType:
                if not isinstance(sample_input, (pd.DataFrame, np.ndarray)):
                    sample_input = model_signature._convert_local_data_to_df(sample_input)

                if isinstance(model, xgboost.Booster):
                    sample_input = xgboost.DMatrix(sample_input)

                target_method = getattr(model, target_method_name, None)
                assert callable(target_method)
                predictions_df = target_method(sample_input)
                return predictions_df

            model_meta = model_meta_api._validate_signature(
                model=model,
                model_meta=model_meta,
                target_methods=target_methods,
                sample_input=sample_input,
                get_prediction_fn=get_prediction,
            )

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        os.makedirs(model_blob_path, exist_ok=True)
        model.save_model(os.path.join(model_blob_path, _XGBModelHandler.MODEL_BLOB_FILE))
        base_meta = model_meta_api._ModelBlobMetadata(
            name=name,
            model_type=_XGBModelHandler.handler_type,
            path=_XGBModelHandler.MODEL_BLOB_FILE,
            options={"xgb_estimator_type": model.__class__.__name__},
        )
        model_meta.models[name] = base_meta
        model_meta._include_if_absent(
            [
                model_meta_api.Dependency(conda_name="scikit-learn", pip_name="scikit-learn"),
                model_meta_api.Dependency(conda_name="xgboost", pip_name="xgboost"),
            ]
        )

        model_meta.cuda_version = kwargs.get("cuda_version", model_meta_api._DEFAULT_CUDA_VERSION)

    @staticmethod
    def _load_model(
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> Union["xgboost.Booster", "xgboost.XGBModel"]:
        import xgboost

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        if not hasattr(model_meta, "models"):
            raise ValueError("Ill model metadata found.")
        model_blobs_metadata = model_meta.models
        if name not in model_blobs_metadata:
            raise ValueError(f"Blob of model {name} does not exist.")
        model_blob_metadata = model_blobs_metadata[name]
        model_blob_filename = model_blob_metadata.path
        xgb_estimator_type = model_blob_metadata.options.get("xgb_estimator_type", None)
        if not xgb_estimator_type or not hasattr(xgboost, xgb_estimator_type):
            raise ValueError("Type of XGB estimator unknown or illegal.")
        m = getattr(xgboost, xgb_estimator_type)()
        m.load_model(os.path.join(model_blob_path, model_blob_filename))

        if kwargs.get("use_gpu", False):
            gpu_params = {"tree_method": "gpu_hist", "predictor": "gpu_predictor"}
            if isinstance(m, xgboost.Booster):
                m.set_param(gpu_params)
            elif isinstance(m, xgboost.XGBModel):
                m.set_params(**gpu_params)

        assert isinstance(m, xgboost.Booster) or isinstance(m, xgboost.XGBModel)
        return m

    @staticmethod
    def _load_as_custom_model(
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> custom_model.CustomModel:
        """Create a custom model class wrap for unified interface when being deployed. The predict method will be
        re-targeted based on target_method metadata.

        Args:
            name: Name of the model.
            model_meta: The model metadata.
            model_blobs_dir_path: Directory path to the whole model.
            kwargs: Options when loading the model.

        Returns:
            The model object as a custom model.
        """
        import xgboost

        from snowflake.ml.model import custom_model

        def _create_custom_model(
            raw_model: Union["xgboost.Booster", "xgboost.XGBModel"],
            model_meta: model_meta_api.ModelMetadata,
        ) -> Type[custom_model.CustomModel]:
            def fn_factory(
                raw_model: Union["xgboost.Booster", "xgboost.XGBModel"],
                signature: model_signature.ModelSignature,
                target_method: str,
            ) -> Callable[[custom_model.CustomModel, pd.DataFrame], pd.DataFrame]:
                @custom_model.inference_api
                def fn(self: custom_model.CustomModel, X: pd.DataFrame) -> pd.DataFrame:
                    if isinstance(raw_model, xgboost.Booster):
                        X = xgboost.DMatrix(X)

                    res = getattr(raw_model, target_method)(X)

                    if isinstance(res, list) and len(res) > 0 and isinstance(res[0], np.ndarray):
                        # In case of multi-output estimators, predict_proba(), decision_function(), etc., functions
                        # return a list of ndarrays. We need to deal them seperately
                        df = numpy_handler.SeqOfNumpyArrayHandler.convert_to_df(res)
                    else:
                        df = pd.DataFrame(res)

                    return model_signature_utils.rename_pandas_df(df, signature.outputs)

                return fn

            type_method_dict = {}
            for target_method_name, sig in model_meta.signatures.items():
                type_method_dict[target_method_name] = fn_factory(raw_model, sig, target_method_name)

            _XGBModel = type(
                "_XGBModel",
                (custom_model.CustomModel,),
                type_method_dict,
            )

            return _XGBModel

        raw_model = _XGBModelHandler._load_model(name, model_meta, model_blobs_dir_path, **kwargs)
        _XGBModel = _create_custom_model(raw_model, model_meta)
        xgb_model = _XGBModel(custom_model.ModelContext())

        return xgb_model



#!/usr/bin/env python3
#
# Copyright (c) 2012-2022 Snowflake Computing Inc. All rights reserved.
#
# TODO(hayu): [SNOW-750748] Rename/split framework/utils to constants.py, extract_args.py, etc.
import inspect
import warnings
from enum import Enum
from typing import Any, Callable, Dict, Iterable, Optional, Union

import numpy as np
import sklearn
from packaging import version

from snowflake import snowpark
from snowflake.ml._internal.exceptions import (
    error_codes,
    error_messages,
    exceptions,
    exceptions as snowml_exceptions,
    modeling_error_messages,
)
from snowflake.snowpark import exceptions as snowpark_exceptions, functions as F
from snowflake.snowpark._internal import utils

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

# numeric states to the corresponding Snowpark functions
NUMERIC_STATE_TO_FUNC_DICT = {
    "count": F.count,
    "max": F.max,
    "mean": F.avg,
    "median": F.median,
    "min": F.min,
    "stddev": F.stddev,
    "stddev_pop": F.stddev_pop,
    "variance": F.variance,
    "var_pop": F.var_pop,
}

# basic states to the corresponding Snowpark functions
BASIC_STATE_TO_FUNC_DICT = {
    "mode": F.mode,
}

# states, as the combination of numeric and basic states,
# to the corresponding Snowpark functions,
# used to convert state strings to utility and SQL functions
STATE_TO_FUNC_DICT = {**NUMERIC_STATE_TO_FUNC_DICT, **BASIC_STATE_TO_FUNC_DICT}


class NumericStatistics(str, Enum):
    COUNT = "count"
    MAX = "max"
    MEAN = "mean"
    MEDIAN = "median"
    MIN = "min"
    STDDEV = "stddev"
    STDDEV_POP = "stddev_pop"
    VARIANCE = "variance"
    VAR_POP = "var_pop"


class BasicStatistics(str, Enum):
    MODE = "mode"


def get_default_args(func: Callable[..., None]) -> Dict[str, Any]:
    signature = inspect.signature(func)
    return {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}


def generate_value_with_prefix(prefix: str) -> str:
    return f"{prefix}{utils.generate_random_alphanumeric()}"


def get_filtered_valid_sklearn_args(
    args: Dict[str, Any],
    default_sklearn_args: Dict[str, Any],
    sklearn_initial_keywords: Optional[Union[str, Iterable[str]]] = None,
    sklearn_unused_keywords: Optional[Union[str, Iterable[str]]] = None,
    snowml_only_keywords: Optional[Union[str, Iterable[str]]] = None,
    sklearn_added_keyword_to_version_dict: Optional[Dict[str, str]] = None,
    sklearn_added_kwarg_value_to_version_dict: Optional[Dict[str, Dict[str, str]]] = None,
    sklearn_deprecated_keyword_to_version_dict: Optional[Dict[str, str]] = None,
    sklearn_removed_keyword_to_version_dict: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    """
    Get valid sklearn keyword arguments with non-default values.

    Validate if the scikit-learn package version meets the requirements.
    This method validates both keywords and values (e.g. OneHotEncoder
    handle_unknown='infrequent_if_exist').

    Args:
        args: Keyword arguments to validate.
        default_sklearn_args: Default sklearn keyword arguments.
        sklearn_initial_keywords: Initial keywords in sklearn.
        sklearn_unused_keywords: Sklearn keywords that are unused in snowml.
        snowml_only_keywords: snowml only keywords not present in sklearn.
        sklearn_added_keyword_to_version_dict: Added keywords mapped to the sklearn versions in which they were added.
        sklearn_added_kwarg_value_to_version_dict: Added keyword argument values mapped to the sklearn versions
            in which they were added.
        sklearn_deprecated_keyword_to_version_dict: Deprecated keywords mapped to the sklearn versions in which
            they were deprecated.
        sklearn_removed_keyword_to_version_dict: Removed keywords mapped to the sklearn versions in which they
            were removed.

    Returns:
        Sklearn keyword arguments.

    Raises:
        SnowflakeMLException: If the input args contains an invalid key.
        SnowflakeMLException: If the scikit-learn package version does not meet the requirements
            for the keyword arguments.
    """
    # get args to be passed to sklearn
    sklearn_args = {}
    for key, val in args.items():
        # initial sklearn keyword
        if sklearn_initial_keywords and (key in sklearn_initial_keywords):
            sklearn_args[key] = val
        # deprecated sklearn keyword
        elif sklearn_deprecated_keyword_to_version_dict and (key in sklearn_deprecated_keyword_to_version_dict):
            sklearn_args[key] = val
        # removed sklearn keyword
        elif sklearn_removed_keyword_to_version_dict and (key in sklearn_removed_keyword_to_version_dict):
            sklearn_args[key] = val
        # added sklearn keyword
        elif sklearn_added_keyword_to_version_dict and (key in sklearn_added_keyword_to_version_dict):
            # If the keyword is not in sklearn, then there is likely a
            # version mismatch. The arg is still added to `sklearn_args`
            # for sklearn version validation.
            if key not in default_sklearn_args:
                sklearn_args[key] = val
            # pass non-default values to sklearn to avoid raising an
            # unexpected argument error even if the argument is not
            # passed to the snowml transformer
            elif val != default_sklearn_args[key]:
                # check if both values are numpy.nan since numpy.nan != numpy.nan
                if isinstance(val, float) and (np.isnan(val) and np.isnan(default_sklearn_args[key])):
                    continue
                sklearn_args[key] = val
        # unused sklearn keyword in snowml
        elif sklearn_unused_keywords and (key in sklearn_unused_keywords):
            continue
        # snowml only keyword
        elif snowml_only_keywords and (key in snowml_only_keywords):
            continue
        # unknown keyword
        else:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(error_messages.UNEXPECTED_KEYWORD.format(key)),
            )

    # validate sklearn version
    sklearn_version = sklearn.__version__
    for key, val in sklearn_args.items():
        # added sklearn keyword
        if (
            sklearn_added_keyword_to_version_dict
            and (key in sklearn_added_keyword_to_version_dict)
            and (version.parse(sklearn_version) < version.parse(sklearn_added_keyword_to_version_dict[key]))
        ):
            required_version = sklearn_added_keyword_to_version_dict[key]
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.DEPENDENCY_VERSION_ERROR,
                original_exception=ImportError(
                    modeling_error_messages.INCOMPATIBLE_NEW_SKLEARN_PARAM.format(
                        key, required_version, sklearn_version
                    )
                ),
            )

        # added keyword argument value
        if (
            sklearn_added_kwarg_value_to_version_dict
            and (key in sklearn_added_kwarg_value_to_version_dict)
            and (isinstance(val, str) and val in sklearn_added_kwarg_value_to_version_dict[key])
            and (version.parse(sklearn_version) < version.parse(sklearn_added_kwarg_value_to_version_dict[key][val]))
        ):
            required_version = sklearn_added_kwarg_value_to_version_dict[key][val]
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.DEPENDENCY_VERSION_ERROR,
                original_exception=ImportError(
                    modeling_error_messages.INCOMPATIBLE_NEW_SKLEARN_PARAM.format(
                        f"{key}={val}", required_version, sklearn_version
                    )
                ),
            )

        # deprecated sklearn keyword
        if (
            sklearn_deprecated_keyword_to_version_dict
            and (key in sklearn_deprecated_keyword_to_version_dict)
            and (version.parse(sklearn_version) >= version.parse(sklearn_deprecated_keyword_to_version_dict[key]))
        ):
            deprecated_version = sklearn_deprecated_keyword_to_version_dict[key]
            msg = f"Incompatible scikit-learn version: '{key}' deprecated since scikit-learn={deprecated_version}.."
            warnings.warn(msg, DeprecationWarning)

        # removed sklearn keyword
        if (
            sklearn_removed_keyword_to_version_dict
            and (key in sklearn_removed_keyword_to_version_dict)
            and (version.parse(sklearn_version) >= version.parse(sklearn_removed_keyword_to_version_dict[key]))
        ):
            removed_version = sklearn_removed_keyword_to_version_dict[key]
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.DEPENDENCY_VERSION_ERROR,
                original_exception=ImportError(
                    modeling_error_messages.REMOVED_SKLEARN_PARAM.format(key, removed_version, sklearn_version)
                ),
            )

    return sklearn_args


def str_to_bool(value: str) -> Union[bool, None]:
    if value is None:
        return None
    elif value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    return None


def _handle_str_bool_type() -> np.vectorize:
    return np.vectorize(str_to_bool)


def to_native_format(obj: Any) -> Any:
    if "XGB" in obj.__class__.__name__:
        return obj.to_xgboost()
    elif "LGBM" in obj.__class__.__name__:
        return obj.to_lightgbm()
    return obj.to_sklearn()


def table_exists(session: snowpark.Session, table_name: str, statement_params: Dict[str, Any]) -> bool:
    try:
        session.table(table_name).limit(0).collect(statement_params=statement_params)
        return True
    except snowpark_exceptions.SnowparkSQLException:
        return False



import dataclasses
import importlib
import os
import sys
import warnings
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime
from types import ModuleType
from typing import Any, Callable, Dict, Generator, List, Optional, Sequence, cast

import cloudpickle
import yaml
from packaging import version

from snowflake.ml._internal import env as snowml_env, env_utils, file_utils
from snowflake.ml.model import (
    _core_requirements,
    _env,
    model_signature,
    type_hints as model_types,
)
from snowflake.ml.model._signatures import snowpark_handler
from snowflake.snowpark import DataFrame as SnowparkDataFrame

MODEL_METADATA_VERSION = 1
_BASIC_DEPENDENCIES = _core_requirements.REQUIREMENTS
_SNOWFLAKE_PKG_NAME = "snowflake"
_SNOWFLAKE_ML_PKG_NAME = f"{_SNOWFLAKE_PKG_NAME}.ml"
_DEFAULT_CUDA_VERSION = "11.7"

Dependency = namedtuple("Dependency", ["conda_name", "pip_name"])


@dataclasses.dataclass
class _ModelBlobMetadata:
    """Dataclass to store metadata of an individual model blob (sub-model) in the packed model.

    Attributes:
        name: The name to refer the sub-model.
        model_type: The type of the model and handler to use.
        path: Path to the picked model file. It is a relative path from the model blob directory.
        artifacts: Optional, used in custom model to show the mapping between artifact name and relative path
            from the model blob directory.
        options: Optional, used for some model specific metadata storage
    """

    name: str
    model_type: str
    path: str
    artifacts: Dict[str, str] = dataclasses.field(default_factory=dict)
    options: Dict[str, str] = dataclasses.field(default_factory=dict)


@contextmanager
def _create_model_metadata(
    *,
    model_dir_path: str,
    name: str,
    model_type: str,
    signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
    metadata: Optional[Dict[str, str]] = None,
    code_paths: Optional[List[str]] = None,
    ext_modules: Optional[List[ModuleType]] = None,
    conda_dependencies: Optional[List[str]] = None,
    pip_requirements: Optional[List[str]] = None,
    python_version: Optional[str] = None,
    **kwargs: Any,
) -> Generator["ModelMetadata", None, None]:
    """Create a generator for model metadata object. Use generator to ensure correct register and unregister for
        cloudpickle.

    Args:
        model_dir_path: Path to the directory containing the model to be packed.
        name: Name of the model.
        model_type: Type of the model.
        signatures: Signatures of the model. If None, it will be inferred after the model meta is created.
            Defaults to None.
        metadata: User provided key-value metadata of the model. Defaults to None.
        code_paths: List of paths to additional codes that needs to be packed with. Defaults to None.
        ext_modules: List of names of modules that need to be pickled with the model. Defaults to None.
        conda_dependencies: List of conda requirements for running the model. Defaults to None.
        pip_requirements: List of pip Python packages requirements for running the model. Defaults to None.
        python_version: A string of python version where model is run. Used for user override. If specified as None,
            current version would be captured. Defaults to None.
        **kwargs: Dict of attributes and values of the metadata. Used when loading from file.

    Raises:
        ValueError: Raised when the code path contains reserved file or directory.

    Yields:
        A model metadata object.
    """
    model_dir_path = os.path.normpath(model_dir_path)
    embed_local_ml_library = kwargs.pop("embed_local_ml_library", False)
    # Use the last one which is loaded first, that is mean, it is loaded from site-packages.
    # We could make sure that user does not overwrite our library with their code follow the same naming.
    snowml_path = list(importlib.import_module(_SNOWFLAKE_ML_PKG_NAME).__path__)[-1]
    if embed_local_ml_library:
        kwargs["local_ml_library_version"] = f"{snowml_env.VERSION}+{file_utils.hash_directory(snowml_path)}"

    model_meta = ModelMetadata(
        name=name,
        metadata=metadata,
        model_type=model_type,
        conda_dependencies=conda_dependencies,
        pip_requirements=pip_requirements,
        python_version=python_version,
        signatures=signatures,
        **kwargs,
    )

    code_dir_path = os.path.join(model_dir_path, ModelMetadata.MODEL_CODE_DIR)
    if embed_local_ml_library or code_paths:
        os.makedirs(code_dir_path, exist_ok=True)

    if embed_local_ml_library:
        snowml_path_in_code = os.path.join(code_dir_path, _SNOWFLAKE_PKG_NAME)
        os.makedirs(snowml_path_in_code, exist_ok=True)
        file_utils.copy_file_or_tree(snowml_path, snowml_path_in_code)

    if code_paths:
        for code_path in code_paths:
            # This part is to prevent users from providing code following our naming and overwrite our code.
            if (
                os.path.isfile(code_path) and os.path.splitext(os.path.basename(code_path))[0] == _SNOWFLAKE_PKG_NAME
            ) or (os.path.isdir(code_path) and os.path.basename(code_path) == _SNOWFLAKE_PKG_NAME):
                raise ValueError("`snowflake` is a reserved name and you cannot contain that into code path.")
            file_utils.copy_file_or_tree(code_path, code_dir_path)

    try:
        imported_modules = []
        if ext_modules:
            registered_modules = cloudpickle.list_registry_pickle_by_value()
            for mod in ext_modules:
                if mod.__name__ not in registered_modules:
                    cloudpickle.register_pickle_by_value(mod)
                    imported_modules.append(mod)
        yield model_meta
        model_meta.save_model_metadata(model_dir_path)
    finally:
        for mod in imported_modules:
            cloudpickle.unregister_pickle_by_value(mod)


def _load_model_metadata(model_dir_path: str) -> "ModelMetadata":
    """Load models for a directory. Model is initially loaded normally. If additional codes are included when packed,
        the code path is added to system path to be imported with highest priority.

    Args:
        model_dir_path: Path to the directory containing the model to be loaded.

    Returns:
        A model metadata object.
    """
    model_dir_path = os.path.normpath(model_dir_path)
    meta = ModelMetadata.load_model_metadata(model_dir_path)
    return meta


def _load_code_path(model_dir_path: str) -> None:
    """Load custom code in the code path into memory.

    Args:
        model_dir_path: Path to the directory containing the model to be loaded.

    """
    model_dir_path = os.path.normpath(model_dir_path)
    code_path = os.path.join(model_dir_path, ModelMetadata.MODEL_CODE_DIR)
    if os.path.exists(code_path):
        if code_path in sys.path:
            sys.path.remove(code_path)
        sys.path.insert(0, code_path)
        module_names = file_utils.get_all_modules(code_path)
        # If the module_name starts with snowflake, then do not replace it.
        # When deploying, we would add them beforehand.
        # When in the local, they should not be added. We already prevent user from overwriting us.
        module_names = [
            module_name
            for module_name in module_names
            if not (module_name.startswith(f"{_SNOWFLAKE_PKG_NAME}.") or module_name == _SNOWFLAKE_PKG_NAME)
        ]
        for module_name in module_names:
            actual_module = sys.modules.pop(module_name, None)
            if actual_module is not None:
                sys.modules[module_name] = importlib.import_module(module_name)

        assert code_path in sys.path
        sys.path.remove(code_path)


class ModelMetadata:
    """Model metadata for Snowflake native model packaged model.

    Attributes:
        name: Name of the model.
        model_type: Type of the model.
        creation_timestamp: Unix timestamp when the model metadata is created.
        python_version: String 'major.minor.patchlevel' showing the python version where the model runs.
        cuda_version: CUDA version to be used, if None then the model cannot be deployed to instance with GPUs.
    """

    MANIFEST_FILE = "MANIFEST"
    ENV_DIR = "env"
    MODEL_CODE_DIR = "code"
    MODEL_METADATA_FILE = "model.yaml"

    def __init__(
        self,
        *,
        name: str,
        model_type: str,
        signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
        metadata: Optional[Dict[str, str]] = None,
        conda_dependencies: Optional[List[str]] = None,
        pip_requirements: Optional[List[str]] = None,
        python_version: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the model metadata. Anything indicates in kwargs has higher priority.

        Args:
            name: Name of the model.
            model_type: Type of the model.
            signatures: A dict mapping from target function name to input and output signatures.
            metadata: User provided key-value metadata of the model. Defaults to None.
            conda_dependencies: List of conda requirements for running the model. Defaults to None.
            pip_requirements: List of pip Python packages requirements for running the model. Defaults to None.
            python_version: A string of python version where model is run. Used for user override. If specified as None,
                current version would be captured. Defaults to None.
            **kwargs: Dict of attributes and values of the metadata. Used when loading from file.

        Raises:
            ValueError: Raised when the user provided version string is invalid.
        """
        self.name = name
        self._signatures = signatures
        self.metadata = metadata
        self.creation_timestamp = str(datetime.utcnow())
        self.model_type = model_type
        self._models: Dict[str, _ModelBlobMetadata] = dict()
        if python_version:
            try:
                self.python_version = str(version.parse(python_version))
                # We might have more check here later.
            except version.InvalidVersion:
                raise ValueError(f"{python_version} is not a valid Python version.")
        else:
            self.python_version = snowml_env.PYTHON_VERSION

        self._conda_dependencies = env_utils.validate_conda_dependency_string_list(
            conda_dependencies if conda_dependencies else []
        )
        self._pip_requirements = env_utils.validate_pip_requirement_string_list(
            pip_requirements if pip_requirements else []
        )
        if "local_ml_library_version" in kwargs:
            self._include_if_absent([Dependency(conda_name=dep, pip_name=dep) for dep in _BASIC_DEPENDENCIES])
        else:
            self._include_if_absent(
                [Dependency(conda_name=dep, pip_name=dep) for dep in _BASIC_DEPENDENCIES + [env_utils._SNOWML_PKG_NAME]]
            )
        self._cuda_version: Optional[str] = None

        self.__dict__.update(kwargs)

    @property
    def pip_requirements(self) -> List[str]:
        """List of pip Python packages requirements for running the model."""
        return list(sorted(map(str, self._pip_requirements)))

    @property
    def conda_dependencies(self) -> List[str]:
        """List of conda channel and dependencies from that to run the model"""
        return sorted(
            f"{chan}::{str(req)}" if chan else str(req)
            for chan, reqs in self._conda_dependencies.items()
            for req in reqs
        )

    def _include_if_absent(self, pkgs: List[Dependency]) -> None:
        conda_reqs_str, pip_reqs_str = tuple(zip(*pkgs))
        pip_reqs = env_utils.validate_pip_requirement_string_list(list(pip_reqs_str))
        conda_reqs = env_utils.validate_conda_dependency_string_list(list(conda_reqs_str))

        for conda_req, pip_req in zip(conda_reqs[env_utils.DEFAULT_CHANNEL_NAME], pip_reqs):
            req_to_add = env_utils.get_local_installed_version_of_pip_package(pip_req)
            req_to_add.name = conda_req.name
            for added_pip_req in self._pip_requirements:
                if added_pip_req.name == pip_req.name:
                    warnings.warn(
                        (
                            f"Basic dependency {conda_req} specified from PIP requirements."
                            + " This may prevent model deploying to Snowflake Warehouse."
                        ),
                        category=UserWarning,
                    )
            try:
                env_utils.append_conda_dependency(
                    self._conda_dependencies, (env_utils.DEFAULT_CHANNEL_NAME, req_to_add)
                )
            except env_utils.DuplicateDependencyError:
                pass
            except env_utils.DuplicateDependencyInMultipleChannelsError:
                warnings.warn(
                    (
                        f"Basic dependency {conda_req.name} specified from non-Snowflake channel."
                        + " This may prevent model deploying to Snowflake Warehouse."
                    ),
                    category=UserWarning,
                )

    @property
    def cuda_version(self) -> Optional[str]:
        return self._cuda_version

    @cuda_version.setter
    def cuda_version(self, _cuda_version: str) -> None:
        if not isinstance(_cuda_version, str):
            raise ValueError("Cannot set CUDA version as a non-str object.")
        if self._cuda_version is None:
            self._cuda_version = _cuda_version
        else:
            if self._cuda_version != _cuda_version:
                raise ValueError(
                    f"Different CUDA version {self._cuda_version} and {_cuda_version} found in the same model!"
                )

    @property
    def signatures(self) -> Dict[str, model_signature.ModelSignature]:
        """Signatures of the model.

        Raises:
            RuntimeError: Raised when the metadata is not ready to save

        Returns:
            Model signatures.
        """
        if self._signatures is None:
            raise RuntimeError("The meta data is not ready to save.")
        return self._signatures

    @property
    def models(self) -> Dict[str, _ModelBlobMetadata]:
        """Dict showing the mapping from sub-models' name to corresponding model blob metadata."""
        return self._models

    def to_dict(self) -> Dict[str, Any]:
        """Serialize to a dictionary.

        Raises:
            RuntimeError: Raised when the metadata is not ready to save

        Returns:
            A dict containing the information of the model metadata.
        """
        if self._signatures is None:
            raise RuntimeError("The meta data is not ready to save.")
        res = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        res["signatures"] = {func_name: sig.to_dict() for func_name, sig in self._signatures.items()}
        res["models"] = {name: dataclasses.asdict(blob_meta) for name, blob_meta in self._models.items()}
        res["pip_requirements"] = self.pip_requirements
        res["conda_dependencies"] = self.conda_dependencies
        res["cuda_version"] = self._cuda_version
        return res

    @classmethod
    def from_dict(cls, model_dict: Dict[str, Any]) -> "ModelMetadata":
        """Deserialize from a dictionary.

        Args:
            model_dict: The dict where metadata is stored.

        Returns:
            A model metadata object created from the given dict.
        """
        model_dict["signatures"] = {
            func_name: model_signature.ModelSignature.from_dict(sig)
            for func_name, sig in model_dict.pop("signatures").items()
        }
        model_dict["_models"] = {
            name: _ModelBlobMetadata(**blob_meta) for name, blob_meta in model_dict.pop("models").items()
        }
        model_dict["_cuda_version"] = model_dict.pop("cuda_version", None)
        return cls(**model_dict)

    def save_model_metadata(self, path: str) -> None:
        """Save the model metadata as a yaml file in the model directory.

        Args:
            path: The path of the directory to write a yaml file in it.
        """
        model_yaml_path = os.path.join(path, ModelMetadata.MODEL_METADATA_FILE)
        with open(model_yaml_path, "w", encoding="utf-8") as out:
            yaml.safe_dump({**self.to_dict(), "version": MODEL_METADATA_VERSION}, stream=out, default_flow_style=False)

        env_dir_path = os.path.join(path, ModelMetadata.ENV_DIR)
        os.makedirs(env_dir_path, exist_ok=True)

        _env.save_conda_env_file(env_dir_path, self._conda_dependencies, self.python_version)
        _env.save_requirements_file(env_dir_path, self._pip_requirements)

    @classmethod
    def load_model_metadata(cls, path: str) -> "ModelMetadata":
        """Load the model metadata from the model metadata yaml file in the model directory.

        Args:
            path: The path of the directory to read the metadata yaml file in it.

        Raises:
            NotImplementedError: raised when version is not found or unsupported in metadata file.

        Returns:
            Loaded model metadata object.
        """
        model_yaml_path = os.path.join(path, ModelMetadata.MODEL_METADATA_FILE)
        with open(model_yaml_path, encoding="utf-8") as f:
            loaded_mata = yaml.safe_load(f.read())

        loaded_mata_version = loaded_mata.pop("version", None)
        if not loaded_mata_version or loaded_mata_version != MODEL_METADATA_VERSION:
            raise NotImplementedError("Unknown or unsupported model metadata file found.")

        meta = ModelMetadata.from_dict(loaded_mata)
        env_dir_path = os.path.join(path, ModelMetadata.ENV_DIR)
        meta._conda_dependencies, python_version = _env.load_conda_env_file(
            os.path.join(env_dir_path, _env._CONDA_ENV_FILE_NAME)
        )
        if python_version:
            meta.python_version = python_version
        meta._pip_requirements = _env.load_requirements_file(os.path.join(env_dir_path, _env._REQUIREMENTS_FILE_NAME))
        return meta


def _is_callable(model: model_types.SupportedModelType, method_name: str) -> bool:
    return callable(getattr(model, method_name, None))


def _validate_signature(
    model: model_types.SupportedRequireSignatureModelType,
    model_meta: ModelMetadata,
    target_methods: Sequence[str],
    sample_input: Optional[model_types.SupportedDataType],
    get_prediction_fn: Callable[[str, model_types.SupportedLocalDataType], model_types.SupportedLocalDataType],
) -> ModelMetadata:
    if model_meta._signatures is not None:
        _validate_target_methods(model, list(model_meta.signatures.keys()))
        return model_meta

    # In this case sample_input should be available, because of the check in save_model.
    assert (
        sample_input is not None
    ), "Model signature and sample input are None at the same time. This should not happen with local model."
    model_meta._signatures = {}
    trunc_sample_input = model_signature._truncate_data(sample_input)
    if isinstance(sample_input, SnowparkDataFrame):
        # Added because of Any from missing stubs.
        trunc_sample_input = cast(SnowparkDataFrame, trunc_sample_input)
        local_sample_input = snowpark_handler.SnowparkDataFrameHandler.convert_to_df(trunc_sample_input)
    else:
        local_sample_input = trunc_sample_input
    for target_method in target_methods:
        predictions_df = get_prediction_fn(target_method, local_sample_input)
        sig = model_signature.infer_signature(sample_input, predictions_df)
        model_meta._signatures[target_method] = sig
    return model_meta


def _get_target_methods(
    model: model_types.SupportedModelType,
    target_methods: Optional[Sequence[str]],
    default_target_methods: Sequence[str],
) -> Sequence[str]:
    if target_methods is None:
        target_methods = [method_name for method_name in default_target_methods if _is_callable(model, method_name)]

    _validate_target_methods(model, target_methods)
    return target_methods


def _validate_target_methods(model: model_types.SupportedModelType, target_methods: Sequence[str]) -> None:
    for method_name in target_methods:
        if method_name not in target_methods:
            raise ValueError(f"Target method {method_name} does not exists.")
        if not _is_callable(model, method_name):
            raise ValueError(f"Target method {method_name} is not callable.")



import textwrap
from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    final,
)

import numpy as np
import numpy.typing as npt

import snowflake.snowpark.types as spt
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)

if TYPE_CHECKING:
    import mlflow
    import torch


class DataType(Enum):
    def __init__(self, value: str, snowpark_type: Type[spt.DataType], numpy_type: npt.DTypeLike) -> None:
        self._value = value
        self._snowpark_type = snowpark_type
        self._numpy_type = numpy_type

    INT8 = ("int8", spt.ByteType, np.int8)
    INT16 = ("int16", spt.ShortType, np.int16)
    INT32 = ("int32", spt.IntegerType, np.int32)
    INT64 = ("int64", spt.LongType, np.int64)

    FLOAT = ("float", spt.FloatType, np.float32)
    DOUBLE = ("double", spt.DoubleType, np.float64)

    UINT8 = ("uint8", spt.ByteType, np.uint8)
    UINT16 = ("uint16", spt.ShortType, np.uint16)
    UINT32 = ("uint32", spt.IntegerType, np.uint32)
    UINT64 = ("uint64", spt.LongType, np.uint64)

    BOOL = ("bool", spt.BooleanType, np.bool_)
    STRING = ("string", spt.StringType, np.str_)
    BYTES = ("bytes", spt.BinaryType, np.bytes_)

    def as_snowpark_type(self) -> spt.DataType:
        """Convert to corresponding Snowpark Type.

        Returns:
            A Snowpark type.
        """
        return self._snowpark_type()

    def __repr__(self) -> str:
        return f"DataType.{self.name}"

    @classmethod
    def from_numpy_type(cls, np_type: npt.DTypeLike) -> "DataType":
        """Translate numpy dtype to DataType for signature definition.

        Args:
            np_type: The numpy dtype.

        Raises:
            SnowflakeMLException: NotImplementedError: Raised when the given numpy type is not supported.

        Returns:
            Corresponding DataType.
        """
        np_to_snowml_type_mapping = {i._numpy_type: i for i in DataType}
        for potential_type in np_to_snowml_type_mapping.keys():
            if np.can_cast(np_type, potential_type, casting="no"):
                # This is used since the same dtype might represented in different ways.
                return np_to_snowml_type_mapping[potential_type]
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.NOT_IMPLEMENTED,
            original_exception=NotImplementedError(f"Type {np_type} is not supported as a DataType."),
        )

    @classmethod
    def from_torch_type(cls, torch_type: "torch.dtype") -> "DataType":
        import torch

        """Translate torch dtype to DataType for signature definition.

        Args:
            torch_type: The torch dtype.

        Returns:
            Corresponding DataType.
        """
        torch_dtype_to_numpy_dtype_mapping = {
            torch.uint8: np.uint8,
            torch.int8: np.int8,
            torch.int16: np.int16,
            torch.int32: np.int32,
            torch.int64: np.int64,
            torch.float32: np.float32,
            torch.float64: np.float64,
            torch.bool: np.bool_,
        }
        return cls.from_numpy_type(torch_dtype_to_numpy_dtype_mapping[torch_type])

    @classmethod
    def from_snowpark_type(cls, snowpark_type: spt.DataType) -> "DataType":
        """Translate snowpark type to DataType for signature definition.

        Args:
            snowpark_type: The snowpark type.

        Raises:
            SnowflakeMLException: NotImplementedError: Raised when the given numpy type is not supported.

        Returns:
            Corresponding DataType.
        """
        if isinstance(snowpark_type, spt.ArrayType):
            actual_sp_type = snowpark_type.element_type
        else:
            actual_sp_type = snowpark_type

        snowpark_to_snowml_type_mapping: Dict[Type[spt.DataType], DataType] = {
            i._snowpark_type: i
            for i in DataType
            # We by default infer as signed integer.
            if i not in [DataType.UINT8, DataType.UINT16, DataType.UINT32, DataType.UINT64]
        }
        for potential_type in snowpark_to_snowml_type_mapping.keys():
            if isinstance(actual_sp_type, potential_type):
                return snowpark_to_snowml_type_mapping[potential_type]
        # Fallback for decimal type.
        if isinstance(snowpark_type, spt.DecimalType):
            if snowpark_type.scale == 0:
                return DataType.INT64
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.NOT_IMPLEMENTED,
            original_exception=NotImplementedError(f"Type {snowpark_type} is not supported as a DataType."),
        )

    def is_same_snowpark_type(self, incoming_snowpark_type: spt.DataType) -> bool:
        """Check if provided snowpark type is the same as Data Type.

        Args:
            incoming_snowpark_type: The snowpark type.

        Raises:
            SnowflakeMLException: NotImplementedError: Raised when the given numpy type is not supported.

        Returns:
            If the provided snowpark type is the same as the DataType.
        """
        # Special handle for Decimal Type.
        if isinstance(incoming_snowpark_type, spt.DecimalType):
            if incoming_snowpark_type.scale == 0:
                return self == DataType.INT64 or self == DataType.UINT64
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_IMPLEMENTED,
                original_exception=NotImplementedError(
                    f"Type {incoming_snowpark_type} is not supported as a DataType."
                ),
            )

        return isinstance(incoming_snowpark_type, self._snowpark_type)


class BaseFeatureSpec(ABC):
    """Abstract Class for specification of a feature."""

    def __init__(self, name: str) -> None:
        self._name = name

    @final
    @property
    def name(self) -> str:
        """Name of the feature."""
        return self._name

    @abstractmethod
    def as_snowpark_type(self) -> spt.DataType:
        """Convert to corresponding Snowpark Type."""
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Serialization"""
        pass

    @classmethod
    @abstractmethod
    def from_dict(self, input_dict: Dict[str, Any]) -> "BaseFeatureSpec":
        """Deserialization"""
        pass


class FeatureSpec(BaseFeatureSpec):
    """Specification of a feature in Snowflake native model packaging."""

    def __init__(
        self,
        name: str,
        dtype: DataType,
        shape: Optional[Tuple[int, ...]] = None,
    ) -> None:
        """Initialize a feature.

        Args:
            name: Name of the feature.
            dtype: Type of the elements in the feature.
            shape: Used to represent scalar feature, 1-d feature list or n-d tensor.
                -1 is used to represent variable length.Defaults to None.

                E.g.
                None: scalar
                (2,): 1d list with fixed len of 2.
                (-1,): 1d list with variable length. Used for ragged tensor representation.
                (d1, d2, d3): 3d tensor.

        Raises:
            SnowflakeMLException: TypeError: Raised when the dtype input type is incorrect.
            SnowflakeMLException: TypeError: Raised when the shape input type is incorrect.
        """
        super().__init__(name=name)

        if not isinstance(dtype, DataType):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_TYPE,
                original_exception=TypeError("dtype should be a model signature datatype."),
            )
        self._dtype = dtype

        if shape and not isinstance(shape, tuple):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_TYPE,
                original_exception=TypeError("Shape should be a tuple if presented."),
            )
        self._shape = shape

    def as_snowpark_type(self) -> spt.DataType:
        result_type = self._dtype.as_snowpark_type()
        if not self._shape:
            return result_type
        for _ in range(len(self._shape)):
            result_type = spt.ArrayType(result_type)
        return result_type

    def as_dtype(self) -> npt.DTypeLike:
        """Convert to corresponding local Type."""
        if not self._shape:
            return self._dtype._numpy_type
        return np.object_

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FeatureSpec):
            return self._name == other._name and self._dtype == other._dtype and self._shape == other._shape
        else:
            return False

    def __repr__(self) -> str:
        shape_str = f", shape={repr(self._shape)}" if self._shape else ""
        return f"FeatureSpec(dtype={repr(self._dtype)}, name={repr(self._name)}{shape_str})"

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the feature group into a dict.

        Returns:
            A dict that serializes the feature group.
        """
        base_dict: Dict[str, Any] = {
            "type": self._dtype.name,
            "name": self._name,
        }
        if self._shape is not None:
            base_dict["shape"] = self._shape
        return base_dict

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "FeatureSpec":
        """Deserialize the feature specification from a dict.

        Args:
            input_dict: The dict containing information of the feature specification.

        Returns:
            A feature specification instance deserialized and created from the dict.
        """
        name = input_dict["name"]
        shape = input_dict.get("shape", None)
        if shape:
            shape = tuple(shape)
        type = DataType[input_dict["type"]]
        return FeatureSpec(name=name, dtype=type, shape=shape)

    @classmethod
    def from_mlflow_spec(
        cls, input_spec: Union["mlflow.types.ColSpec", "mlflow.types.TensorSpec"], feature_name: str
    ) -> "FeatureSpec":
        import mlflow

        if isinstance(input_spec, mlflow.types.ColSpec):
            name = input_spec.name
            if name is None:
                name = feature_name
            return FeatureSpec(name=name, dtype=DataType.from_numpy_type(input_spec.type.to_numpy()))
        elif isinstance(input_spec, mlflow.types.TensorSpec):
            if len(input_spec.shape) == 1:
                shape = None
            else:
                shape = tuple(input_spec.shape[1:])

            name = input_spec.name
            if name is None:
                name = feature_name
            return FeatureSpec(name=name, dtype=DataType.from_numpy_type(input_spec.type), shape=shape)
        else:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_IMPLEMENTED,
                original_exception=NotImplementedError(f"MLFlow schema type {type(input_spec)} is not supported."),
            )


class FeatureGroupSpec(BaseFeatureSpec):
    """Specification of a group of features in Snowflake native model packaging."""

    def __init__(self, name: str, specs: List[FeatureSpec]) -> None:
        """Initialize a feature group.

        Args:
            name: Name of the feature group.
            specs: A list of feature specifications that composes the group. All children feature specs have to have
                name. And all of them should have the same type.
        """
        super().__init__(name=name)
        self._specs = specs
        self._validate()

    def _validate(self) -> None:
        if len(self._specs) == 0:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT, original_exception=ValueError("No children feature specs.")
            )
        # each has to have name, and same type
        if not all(s._name is not None for s in self._specs):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError("All children feature specs have to have name."),
            )
        if not (all(s._shape is None for s in self._specs) or all(s._shape is not None for s in self._specs)):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError("All children feature specs have to have same shape."),
            )
        first_type = self._specs[0]._dtype
        if not all(s._dtype == first_type for s in self._specs):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError("All children feature specs have to have same type."),
            )

    def as_snowpark_type(self) -> spt.DataType:
        first_type = self._specs[0].as_snowpark_type()
        return spt.MapType(spt.StringType(), first_type)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FeatureGroupSpec):
            return self._specs == other._specs
        else:
            return False

    def __repr__(self) -> str:
        spec_strs = ",\n\t\t".join(repr(spec) for spec in self._specs)
        return textwrap.dedent(
            f"""FeatureGroupSpec(
                name={repr(self._name)},
                specs=[
                    {spec_strs}
                ]
            )
            """
        )

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the feature group into a dict.

        Returns:
            A dict that serializes the feature group.
        """
        return {"feature_group": {"name": self._name, "specs": [s.to_dict() for s in self._specs]}}

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "FeatureGroupSpec":
        """Deserialize the feature group from a dict.

        Args:
            input_dict: The dict containing information of the feature group.

        Returns:
            A feature group instance deserialized and created from the dict.
        """
        specs = []
        for e in input_dict["feature_group"]["specs"]:
            spec = FeatureSpec.from_dict(e)
            specs.append(spec)
        return FeatureGroupSpec(name=input_dict["feature_group"]["name"], specs=specs)


class ModelSignature:
    """Signature of a model that specifies the input and output of a model."""

    def __init__(self, inputs: Sequence[BaseFeatureSpec], outputs: Sequence[BaseFeatureSpec]) -> None:
        """Initialize a model signature

        Args:
            inputs: A sequence of feature specifications and feature group specifications that will compose the
                input of the model.
            outputs: A sequence of feature specifications and feature group specifications that will compose the
                output of the model.
        """
        self._inputs = inputs
        self._outputs = outputs

    @property
    def inputs(self) -> Sequence[BaseFeatureSpec]:
        """Inputs of the model, containing a sequence of feature specifications and feature group specifications."""
        return self._inputs

    @property
    def outputs(self) -> Sequence[BaseFeatureSpec]:
        """Outputs of the model, containing a sequence of feature specifications and feature group specifications."""
        return self._outputs

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ModelSignature):
            return self._inputs == other._inputs and self._outputs == other._outputs
        else:
            return False

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict to represent the whole signature.

        Returns:
            A dict that serializes the signature.
        """

        return {
            "inputs": [spec.to_dict() for spec in self._inputs],
            "outputs": [spec.to_dict() for spec in self._outputs],
        }

    @classmethod
    def from_dict(cls, loaded: Dict[str, Any]) -> "ModelSignature":
        """Create a signature given the dict containing specifications of children features and feature groups.

        Args:
            loaded: The dict to be deserialized.

        Returns:
            A signature deserialized and created from the dict.
        """
        sig_outs = loaded["outputs"]
        sig_inputs = loaded["inputs"]

        deserialize_spec: Callable[[Dict[str, Any]], BaseFeatureSpec] = (
            lambda sig_spec: FeatureGroupSpec.from_dict(sig_spec)
            if "feature_group" in sig_spec
            else FeatureSpec.from_dict(sig_spec)
        )

        return ModelSignature(
            inputs=[deserialize_spec(s) for s in sig_inputs], outputs=[deserialize_spec(s) for s in sig_outs]
        )

    def __repr__(self) -> str:
        inputs_spec_strs = ",\n\t\t".join(repr(spec) for spec in self._inputs)
        outputs_spec_strs = ",\n\t\t".join(repr(spec) for spec in self._outputs)
        return textwrap.dedent(
            f"""ModelSignature(
                    inputs=[
                        {inputs_spec_strs}
                    ],
                    outputs=[
                        {outputs_spec_strs}
                    ]
                )"""
        )

    @classmethod
    def from_mlflow_sig(cls, mlflow_sig: "mlflow.models.ModelSignature") -> "ModelSignature":
        return ModelSignature(
            inputs=[
                FeatureSpec.from_mlflow_spec(spec, f"input_feature_{idx}") for idx, spec in enumerate(mlflow_sig.inputs)
            ],
            outputs=[
                FeatureSpec.from_mlflow_spec(spec, f"output_feature_{idx}")
                for idx, spec in enumerate(mlflow_sig.outputs)
            ],
        )
